"""
\********************************************************************************
* Copyright (c) 2023 the Qrisp authors
*
* This program and the accompanying materials are made available under the
* terms of the Eclipse Public License 2.0 which is available at
* http://www.eclipse.org/legal/epl-2.0.
*
* This Source Code may also be made available under the following Secondary
* Licenses when the conditions for such availability set forth in the Eclipse
* Public License, v. 2.0 are satisfied: GNU General Public License, version 2
* with the GNU Classpath Exception which is
* available at https://www.gnu.org/software/classpath/license.html.
*
* SPDX-License-Identifier: EPL-2.0 OR GPL-2.0 WITH Classpath-exception-2.0
********************************************************************************/
"""

import numpy as np
from qrisp import XGate, PGate, convert_to_qb_list, x, invert, cx, mcx, Qubit
from qrisp.qtypes import QuantumBool, QuantumVariable
from qrisp.mcx_algs.circuit_library import reduced_maslov_qc, margolus_qc, reduced_margolus_qc
from qrisp.mcx_algs.gidney import GidneyLogicalAND
from qrisp.mcx_algs.jones import jones_toffoli



# Ancilla supported multi controlled X with logarithmic depth based on
# https://www.iccs-meeting.org/archive/iccs2022/papers/133530169.pdf
def balauca_mcx(input_qubits, target, ctrl_state=None, phase=None):
    hybrid_mcx(
        input_qubits, target, ctrl_state=ctrl_state, phase=phase, num_ancilla=np.inf
    )


# Hybrid algorithm of yong and balauca with customizable ancilla qubit count.
# Performs several balauca layers and cancels the recursion with yong.
def hybrid_mcx(
        input_qubits,
        target,
        ctrl_state=None,
        phase=None,
        num_ancilla=np.inf,
        num_dirty_ancilla=0,
        use_mcm=False,
):
    """
    Function to dynamically generate mcx gates for a given amount of ancilla qubits.

    Parameters
    ----------
    input_qubits : list[Qubit]
        The Qubits to control on.
    target : Qubit
        The Qubit to target.
    ctrl_state : str, optional
        The control state to activate the X Gate on. The default is "11...".
    phase : float or sympy.Symbol, optional
        If given, this function performs a mcp instead of a mcx. The default is None.
    num_ancilla : int, optional
        The amount of ancillae this function is allowed to use. The default is np.inf.
    count_yong : bool, optional
        If set to False, the ancilla that is used by the yong recursion termination is
        not counted (because it can be dirty). The default is True.

    Returns
    -------
    None.

    """

    input_qubits = list(input_qubits)
    for i in range(len(input_qubits)):
        if isinstance(input_qubits[i], QuantumBool):
            input_qubits[i] = input_qubits[i][0]

    if isinstance(target, Qubit):
        target = [target]
    elif isinstance(target, QuantumVariable):
        target = list(target)

    if ctrl_state is not None:
        for i in range(len(input_qubits)):
            input_qubits[i].ctrl_state = ctrl_state[i]

    

    qs = target[0].qs()

    if len(input_qubits) <= 2 + int(not use_mcm) or num_ancilla == 0:
        ctrl_state = ""
        for qb in input_qubits:
            if hasattr(qb, "ctrl_state"):
                ctrl_state += qb.ctrl_state
            else:
                ctrl_state += "1"

        if len(input_qubits) == 2:
            if phase is None:
                
                if use_mcm:
                    jones_toffoli(input_qubits, target, ctrl_state = ctrl_state)
                else:
                    qs.append(
                        XGate().control(2, ctrl_state=ctrl_state, method="gray"),
                        input_qubits + target,
                    )
            else:
                qs.append(
                    PGate(phase).control(2, ctrl_state=ctrl_state),
                    input_qubits + target,
                )

        elif num_dirty_ancilla and phase is None:
            balauca_dirty(
                input_qubits, target, k=num_dirty_ancilla, ctrl_state=ctrl_state
            )

        else:
            if phase is None:
                qs.append(
                    XGate().control(
                        len(input_qubits), ctrl_state=ctrl_state, method="gray"
                    ),
                    input_qubits + target,
                )
            else:
                qs.append(
                    PGate(phase).control(len(input_qubits), ctrl_state=ctrl_state),
                    input_qubits + target,
                )
        return

    structure = structure_decider(len(input_qubits), num_ancilla)  # [::-1]
    layer_input = []
    layer_structure = []
    layer_output = []
    remainder = list(input_qubits)

    for i in range(len(structure)):
        if not structure[0] <= len(remainder):
            break

        layer_output.append(QuantumBool(name="balauca_anc*", qs = input_qubits[0].qs()))

        for j in range(structure[0]):
            layer_input.append(remainder.pop(0))

        layer_structure.append(structure.pop(0))

    balauca_layer(layer_input, layer_output, structure=layer_structure, invert=False, use_mcm=use_mcm)

    hybrid_mcx(
        layer_output + remainder,
        target,
        num_ancilla=num_ancilla - len(layer_output),
        num_dirty_ancilla=num_dirty_ancilla,
        phase=phase,
        use_mcm = use_mcm
    )

    balauca_layer(layer_input, layer_output, structure=layer_structure, invert=True, use_mcm=use_mcm)

    [qbl.delete() for qbl in layer_output]

    if ctrl_state is not None:
        for i in range(len(input_qubits)):
            del input_qubits[i].ctrl_state

    return


def balauca_layer(input_qubits, output_qubits, structure, invert=False, use_mcm = False):
    if not output_qubits:
        return

    qs = output_qubits[0].qs()
    input_qubits = list(input_qubits)

    counter = 0
    for i in range(len(output_qubits)):
        if structure[i] == 3:
            ctrl_qubits = [
                input_qubits[counter],
                input_qubits[counter + 1],
                input_qubits[counter + 2],
            ]
            counter += 3

            ctrl_state = ""
            for qb in ctrl_qubits:
                if hasattr(qb, "ctrl_state"):
                    ctrl_state += qb.ctrl_state
                else:
                    ctrl_state += "1"

            gate = XGate().control(3, method="gray_pt", ctrl_state=ctrl_state)

            if invert:
                gate = gate.inverse()

            if isinstance(output_qubits[i], Qubit):
                target = output_qubits[i]
            else:
                target = output_qubits[i][0]
                
            qs.append(gate, ctrl_qubits + [target])
            
        else:
            ctrl_qubits = [input_qubits[counter], input_qubits[counter + 1]]
            counter += 2

            ctrl_state = ""
            for qb in ctrl_qubits:
                if hasattr(qb, "ctrl_state"):
                    ctrl_state += qb.ctrl_state
                else:
                    ctrl_state += "1"

            if use_mcm:
                gate = GidneyLogicalAND(ctrl_state=ctrl_state)
            else:
                gate = XGate().control(2, method="gray_pt", ctrl_state=ctrl_state)

            if invert:
                gate = gate.inverse()
            
            if isinstance(output_qubits[i], Qubit):
                target = output_qubits[i]
            else:
                target = output_qubits[i][0]
                
            qs.append(gate, ctrl_qubits + [target])

    return


def structure_decider(n, k):
    # Each element of a Balauca layer reduces the amount of
    # Control qubits for the next layer either by 1 (margolous gate)
    # or by 2 (phase toleratn maslov gate). Ideally, we reduce the amount
    # of control qubits to 3, so we can cancel the recursion
    # and deploy a Toffoli.

    # If we have n controls, p = n - 2 is the amount of reductions,
    # that need to be performed.

    # If we have k ancillae at our disposal, we need to satisfy:

    # p = 2*triple_mcx + (k-triple_mcx_count) = k + triple_mcx_count

    # <=> triple_mcx_count = n - 2 - k

    triple_mcx_count = n - 2 - k

    if triple_mcx_count <= 0:
        return n // 2 * [2]
    elif triple_mcx_count > k:
        return k * [3]
    else:
        return triple_mcx_count * [3] + (k - triple_mcx_count) * [2]


def balauca_dirty(control, target, k, dirty_ancillae=None, ctrl_state=None):
    control = convert_to_qb_list(control)
    target = convert_to_qb_list(target)
    qs = target[0].qs()

    n = len(control)

    k = min((n - 2) // 2 + 1, k)

    if k <= 1:
        qs.append(
            XGate().control(n, ctrl_state=ctrl_state, method="gray"),
            list(control) + [target],
        )
        return

    if ctrl_state is not None:
        for i in range(len(ctrl_state)):
            if ctrl_state[i] == "0":
                x(control[i])

    def reduced_maslov(control, target):
        qs = target.qs()

        qs.append(reduced_maslov_qc.to_gate("reduced_maslov"), control + [target])
        # qs.append(XGate().control(2), control + [target])

    m_1 = int(np.ceil((n - 2 * (k - 1)) / 2))
    m_2 = int(np.floor((n - 2 * (k - 1)) / 2))

    dirty_ancilla_qbls = []
    if dirty_ancillae is None:
        dirty_ancilla_qbls = [
            QuantumBool(name="balauca_dirty*", qs=qs) for i in range(k)
        ]
        dirty_ancillae = [qbl[0] for qbl in dirty_ancilla_qbls]

    upper_block_qubits = control[:m_1] + [dirty_ancillae[0]]
    lower_block_qubits = control[-m_2:] + [dirty_ancillae[-1]]

    def balauca_dirty_helper(control, target, ancillae):
        if len(ancillae) == 0:
            vchain_2_dirty(control, target, dirty_ancillae=lower_block_qubits)
            # mcx(control, target)
            return

        reduced_maslov([ancillae[-1]] + control[-2:], target)

        balauca_dirty_helper(control[:-2], ancillae[-1], ancillae[:-1])

        with invert():
            reduced_maslov([ancillae[-1]] + control[-2:], target)

    vchain_2_dirty(lower_block_qubits, target, dirty_ancillae=upper_block_qubits)

    balauca_dirty_helper(control[:-m_2], dirty_ancillae[-1], dirty_ancillae[:-1])

    vchain_2_dirty(lower_block_qubits, target, dirty_ancillae=upper_block_qubits)

    balauca_dirty_helper(control[:-m_2], dirty_ancillae[-1], dirty_ancillae[:-1])

    [qbl.delete() for qbl in dirty_ancilla_qbls]

    if ctrl_state is not None:
        for i in range(len(ctrl_state)):
            if ctrl_state[i] == "0":
                x(control[i])



def vchain_2_dirty(control, target, dirty_ancillae=None):
    control = convert_to_qb_list(control)
    target = convert_to_qb_list(target)

    if len(control) == 1:
        cx(control, target)
        return
    elif len(control) == 2:
        mcx(control, target, method="gray")

    n = len(control)
    k = n - 2

    dirty_ancilla_qbls = []
    if dirty_ancillae is None:
        dirty_ancilla_qbls = [QuantumBool(name="vchain_2_dirty*") for i in range(k)]
        dirty_ancillae = [qbl[0] for qbl in dirty_ancilla_qbls]

    def reduced_margolus(control, target):
        qs = target.qs()

        control = list(control)
        qs.append(reduced_margolus_qc.to_gate("reduced_margolus"), control + [target])
        # qs.append(XGate().control(2), control + [target])

    def margolus(control, target):
        qs = target.qs()

        control = list(control)
        qs.append(margolus_qc.to_gate("margolus"), control + [target])
        # qs.append(XGate().control(2), control + [target])

    control_temp_list = list(control)
    dirty_ancillae_temp_list = list(dirty_ancillae)
    qubit_list = []

    qubit_list.append(control_temp_list.pop(0))
    qubit_list.append(control_temp_list.pop(0))

    while control_temp_list:
        qubit_list.append(dirty_ancillae_temp_list.pop(0))
        qubit_list.append(control_temp_list.pop(0))

    qubit_list.append(target[0])

    m = len(qubit_list)

    for i in range(k):
        if i == k - 1:
            margolus(
                [qubit_list[m - 5 - 2 * i], qubit_list[m - 4 - 2 * i]],
                qubit_list[m - 3 - 2 * i],
            )
        else:
            reduced_margolus(
                [qubit_list[m - 5 - 2 * i], qubit_list[m - 4 - 2 * i]],
                qubit_list[m - 3 - 2 * i],
            )

    for i in range(k):
        if i == k - 1:
            mcx(
                [qubit_list[2 * i + 2], qubit_list[2 * i + 3]], qubit_list[2 * i + 4], method = "gray"
            )
        else:
            with invert():
                reduced_margolus(
                    [qubit_list[2 * i + 2], qubit_list[2 * i + 3]],
                    qubit_list[2 * i + 4],
                )

    for i in range(k):
        if i == k - 1:
            with invert():
                margolus(
                    [qubit_list[m - 5 - 2 * i], qubit_list[m - 4 - 2 * i]],
                    qubit_list[m - 3 - 2 * i],
                )
        else:
            reduced_margolus(
                [qubit_list[m - 5 - 2 * i], qubit_list[m - 4 - 2 * i]],
                qubit_list[m - 3 - 2 * i],
            )

    for i in range(k):
        if i == k - 1:
            with invert():
                mcx(
                    [qubit_list[2 * i + 2], qubit_list[2 * i + 3]],
                    qubit_list[2 * i + 4], method = "gray"
                )
        else:
            with invert():
                reduced_margolus(
                    [qubit_list[2 * i + 2], qubit_list[2 * i + 3]],
                    qubit_list[2 * i + 4],
                )

    [qbl.delete() for qbl in dirty_ancilla_qbls]


# just write a replacement_routine and throw out not relevant sets

from qrisp import rz, x, cx
import numpy as np
import copy
from qrisp.qiro.qiroproblems.qiro_utils import * 

def create_maxSat_replacement_routine( res, oldproblem, solutions, exclusions):
    
    """
    Creates a replacement routine for the problem structure, i.e. defines the replacement rules. 
    See the original paper for description of the update rules

    Parameters
    ----------
    res : dict
        Result dictionary of initial QAOA optimization procedure.
    oldproblem : List
        The clauses defining the problem instance.
    solutions : List
        Qubits which have been found to be positive correlated, i.e. part of the problem solution.
    exclusions : List
        Qubits which have been found to be negative correlated, i.e. not part of the problem solution, or contradict solution qubits in accordance to the update rules.  

    Returns
    -------
    new_problem: list
        Updated clauses of the problem instance.
    solutions : List
        Updated set of solutions to the problem.
    sign : Int
        The sign of the correlation.
    exclusions : List
        Updated set of exclusions to the problem.



    
    """
    
    numVars = oldproblem[0]
    clauses = copy.deepcopy(oldproblem[1])

    # FOR SINGLE QUBIT CORRELATIONS
    # make -1 here for consistency in the find max
    orig_nodes = [i - 1 for i in list(range(1,numVars + 1)) if not i in exclusions]
    combinations = []

    # add check for solution nodes      
    for index1 in range(len(orig_nodes)-1):
        for index2 in range(index1, len(orig_nodes)):
            combinations.append((orig_nodes[index1],orig_nodes[index2]))
    

    max_item, sign = find_max(orig_nodes, combinations, res, solutions)

    # we just directly remove clauses from clauses
    if isinstance(max_item, int):
        
        max_item += 1

        if sign > 0:
            for item in clauses: 
                # if negation in item --> remove it
                if -1 * max_item in item:
                    clauses.remove(item)
            solutions.append(max_item)
            exclusions.append(max_item)

        elif sign < 0:
            for item in clauses:
                # if number in item --> remove it
                if max_item in item:
                    clauses.remove(item)
            exclusions.append(max_item)

    else:
        max_item[0] += 1 
        max_item[1] += 1 
        if sign > 0:
            for item in clauses:
                # replace with pos. correlated number if its in an item
                if max_item[1] in item:
                    temp = item.index[max_item[1]]
                    item[temp] = max_item[0]
                if -1* max_item[1] in item:
                    temp = item.index[-1* max_item[1]]
                    item[temp] = -1* max_item[0]
                exclusions.append(max_item[1])
                    

        elif sign < 0:
            for item in clauses:
                # replace with neg. correlated number if its in an item
                if max_item[1] in item:
                    temp = item.index[max_item[1]]
                    item[temp] = -1 * max_item[0]
                if -1* max_item[1] in item:
                    temp = item.index[ -1  * max_item[1]]
                    item[temp] = max_item[0]
                exclusions.append(max_item[1])

    # create sign list somewhere?
    return [numVars, clauses], solutions, sign, exclusions


def create_maxSat_cost_operator_reduced(problem, solutions=[]):

    """
    |  Implementation for MaxSat Cost-Operator, in accordance to the original QAOA-paper.
    |  Adjusted for QIRO to also respect found solutions to the problem
    
    Parameters
    ----------
    clauses : List(Lists)
        Clauses to be considered for MaxSat. Should be given as a list of lists and 1-indexed instead 0-indexed, see example
    solutions : List
        Variables that were found to be a part of the solution.
    Returns
    -------
    cost_operator : function
        the Operator applied to the circuit-QuantumVariable

    Examples
    --------

    >>> clauses11 = (6, [[1,2,-3], [1,4,-6], [4,5,6], [1,3,-4], [2,4,5], [1,3,5], [-2,-3,6]])
    
    |  Explanation: 
    |  First entry of tuple is the number of variables, second is the clauses
    |  Clause [1, 2, -4] is fulfilled by the QuantumStates "1100", "1110" 


    * if the sign is positive : the index has of the QuantumVariable to be "1", if negative sign the index has to be "0".
    * indices not mentioned can be "0" or "1" (the third integer in this case).
    * start with 1 in your clauses! because ``int 0`` has no sign and this is problematic. 
    * we want to keep the mathematical formulation, so no handing over strings!

    """
    
    clauses = problem[1]
    satlen = len(clauses[0])

    def maxSatcostEmbed(qv , gamma):
        import itertools
        #clauses = clauses
        
        qc = qv

        for numClause in range(len(clauses)):
            #go through clauses
            clause = clauses[numClause]
        
            for lenofCombination in range(1,satlen+1):
                # get all combinations to assign rz, rzz, rzzz, rzzzzzz....
                combinations = list(itertools.combinations(clause, lenofCombination))
                #print(combinations)
                
                #len == 1: just rz
                if lenofCombination == 1:
                    for index in range(len(combinations)):
                        if  abs( combinations[index][0])-1 not in solutions:
                            # sign for rz mixer
                            signu = - np.sign(combinations[index][0])
                            # always have the "-1" at the end since clauses are not initiated with 0, see above
                            #print(abs( combinations[index][0] - 1 ))
                            rz(signu *  gamma/8, qc[ abs( combinations[index][0]) -1 ] )

                else:
                    #for all combinations of this length 
                    for index in range(len(combinations)):
                        signu = 1

                        # up to final value in combination perform cx gates --> set up the rzz, rzzz, rzzzz ... gates
                        for index2 in range(lenofCombination-1):
                            # (also remember the sign)
                            signu *= - np.sign(combinations[index][index2])
                            cx(qc[abs( combinations[index][index2] ) -1], qc[abs( combinations[index][index2+1] ) -1])

                        signu *= np.sign(combinations[index][lenofCombination-1])
                        # finalize rzz, rzzz, rzzzz ... gates
                        rz(signu*gamma/8, qc[abs( combinations[index][lenofCombination-1] ) -1])
                        
                        # and cnot gates back 
                        for index2 in reversed(range(lenofCombination-1)):
                            cx(  qc[abs(combinations[index][index2] ) -1], qc[abs(combinations[index][index2+1] ) -1 ]) 

        return qc
    
    return maxSatcostEmbed





"""
\********************************************************************************
* Copyright (c) 2023 the Qrisp authors
*
* This program and the accompanying materials are made available under the
* terms of the Eclipse Public License 2.0 which is available at
* http://www.eclipse.org/legal/epl-2.0.
*
* This Source Code may also be made available under the following Secondary
* Licenses when the conditions for such availability set forth in the Eclipse
* Public License, v. 2.0 are satisfied: GNU General Public License, version 2
* with the GNU Classpath Exception which is
* available at https://www.gnu.org/software/classpath/license.html.
*
* SPDX-License-Identifier: EPL-2.0 OR GPL-2.0 WITH Classpath-exception-2.0
********************************************************************************/
"""

import numpy as np

from qrisp import QuantumSession, QuantumVariable, QuantumCircuit

from qrisp.mcx_algs.circuit_library import reduced_margolus_qc, margolus_qc, maslov_qc, reduced_maslov_qc

# Function to synthesize a phase tolerant multi controlled X gate
def pt_multi_cx(n, reduced=False):
    res = QuantumCircuit(n + 1)

    if n == 1:
        res.cx(0, 1)

    # The special cases n = 2 and n = 3 are handled as described in
    # https://arxiv.org/pdf/1508.03273.pdf
    elif n == 2:
        if reduced:
            res.append(reduced_margolus_qc.to_gate("reduced margolus"), res.qubits)
        else:
            res.append(margolus_qc.to_gate("margolus"), res.qubits)

    elif n == 3:
        if reduced:
            res.append(reduced_maslov_qc.to_gate("reduced maslov"), res.qubits)
        else:
            res.append(maslov_qc.to_gate("maslov"), res.qubits)
    else:
        input_qv = QuantumVariable(n)
        output_qv = QuantumVariable(1, qs=input_qv.qs)

        from qrisp import TruthTable
        tt = TruthTable([(2 ** (n) - 1) * "0" + "1"])

        tt.q_synth(input_qv, output_qv, method="gray_pt")

        res = input_qv.qs.copy()

    return res.to_gate(f"pt{n}cx")

def gray_pt_mcx(n, ctrl_state):
    input_qv = QuantumVariable(n)
    output_qv = QuantumVariable(1, qs=input_qv.qs)

    tt_str = 2 ** n * ["0"]
    tt_str[int(ctrl_state[::-1], 2)] = "1"
    from qrisp import TruthTable
    tt = TruthTable([tt_str])

    tt.q_synth(input_qv, output_qv, method="gray_pt")

    res = input_qv.qs.copy()
    return res.to_gate(f"pt{n}cx")



"""
\********************************************************************************
* Copyright (c) 2023 the Qrisp authors
*
* This program and the accompanying materials are made available under the
* terms of the Eclipse Public License 2.0 which is available at
* http://www.eclipse.org/legal/epl-2.0.
*
* This Source Code may also be made available under the following Secondary
* Licenses when the conditions for such availability set forth in the Eclipse
* Public License, v. 2.0 are satisfied: GNU General Public License, version 2
* with the GNU Classpath Exception which is
* available at https://www.gnu.org/software/classpath/license.html.
*
* SPDX-License-Identifier: EPL-2.0 OR GPL-2.0 WITH Classpath-exception-2.0
********************************************************************************/
"""

from qrisp.arithmetic.adders.gidney.cq_gidney_adder import *
from qrisp.arithmetic.adders.gidney.qq_gidney_adder import *
from qrisp.arithmetic.adders.adder_tools import ammend_inpl_adder
from qrisp.environments import custom_control

def gidney_adder(a, b, c_in = None, c_out = None):
    """
    In-place adder function based on `this paper <https://arxiv.org/abs/1709.06648>`_
    performs the addition
    
    ::
        
        b += a
    

    Parameters
    ----------
    a : int or QuantumVariable or list[Qubit]
        The value that should be added.
    b : QuantumVariable or list[Qubit]
        The value that should be modified in the in-place addition.
    c_in : Qubit, optional
        An optional carry in value. The default is None.
    c_out : Qubit, optional
        An optional carry out value. The default is None.

    Examples
    --------
    
    We add two integers:
        
    >>> from qrisp import QuantumFloat, gidney_adder
    >>> a = QuantumFloat(4)
    >>> b = QuantumFloat(4)
    >>> a[:] = 4
    >>> b[:] = 5
    >>> gidney_adder(a,b)
    >>> print(b)
    {9: 1.0}

    """
    
    if isinstance(a, (int, str)):
        return custom_control(cq_gidney_adder)(a, b, c_in = c_in, c_out = c_out)
    else:
        return qq_gidney_adder(a, b, c_in = c_in, c_out = c_out)
        
temp = gidney_adder.__doc__
gidney_adder = ammend_inpl_adder(gidney_adder, ammend_cl_int = False)
gidney_adder.__doc__ = temp


# -*- coding: utf-8 -*-
"""
Created on Tue Nov 21 11:22:37 2023

@author: sea
"""

import numpy as np

from qrisp.core import x, cx
from qrisp.qtypes import QuantumVariable, QuantumBool
from qrisp.environments import invert
from qrisp.misc import redirect_qfunction
from qrisp.arithmetic.adders.gidney import gidney_adder
from qrisp.arithmetic.adders.qcla.quantum_quantum.qq_carry_path import qq_calc_carry
from qrisp.arithmetic.adders.incrementation import lin_incr


def qq_sum_path(a, b, c, R):
    
    if R == 1:
        # If R = 1, we are in the case of Drapers QCLA:
        # https://arxiv.org/abs/quant-ph/0406142   
        # We can use the formula
        # S = A (+) B (+) C
        cx(c[:-1], b[1:])
        cx(a, b[:len(a)])
    else:
        i = 0
        # Execute addition using the corresponding carry values
        for i in range(len(c)+1):
            
            # Determine the radix qubits to perform the addition on
            a_block = a[R*i:R*(i+1)]
            b_block = b[R*i:R*(i+1)]
            
            # If a_block is empty we break the loop.
            # This doesn't necessarily imply that b_block is empty because
            # len(b) could be larger then len(a). In this case we execute an
            # increment block after this loop
            if not len(a_block):
                i = i-1
                break
            
            # If a_block and b_block have different sizes, we create a padding
            # variable to fill up the remaining entries
            if not len(a_block) == len(b_block):
                padding_var = QuantumVariable(len(b_block) - len(a_block))
                a_block = a_block + list(padding_var)
            # Perform Cuccarro addition
            if i == 0:
                # cuccaro_procedure(a.qs, a_block, b_block)
                
                gidney_adder(a_block, b_block)
            else:
                
                # cuccaro_procedure(a.qs, a_block, b_block, carry_in = c[i-1])
                gidney_adder(a_block, b_block, c[i-1])
            
            # Delete carry var
            try:
                padding_var.delete(verify = bool(verify_manual_uncomputations[0]))
            except NameError:
                pass
        
        
        # This loop treats the case that len(a) != len(b)
        # In this case we need to perform the increment function on the remaining
        # qubits of b (incase the carry is True)
        #We start at the index the last loop finished at
        for j in range(i+1, len(b)//R+1)[::-1]:
            
            b_block = b[R*j:R*(j+1)]
            
            if not len(b_block):
                continue
            
            #Perform incrementation function
            if R*j == 0:
                lin_incr(b_block, c_out = c[j])
            else:
                lin_incr(b_block, c_in = c[j-1], c_out = c[j])
            
# This function is a version of the sum path that requires no uncomputation after
# This can be achieved because the the carry_out of the i-th Gidney-Adder
# is used to uncompute the carry_in values of the i+1-th Gidney-adder.

# This however creates another problem: The i+1-th Gidney-adder requires it's carry-in
# at the end of it's computation. Therefore the i-th Gidney adder has to wait until
# the i+1-th finished. This basically reintroduces linear depth.
# BUT
# Not linear T-depth. This is because the right "wing" of the V-shaped part of the
# Gidney-Adder has T-depth 0 regardless of the input size. Since it is only these
# parts that need to wait, the overall T-depth is still logarithmic.

# To understand this function it can be helpfull to first understand the regular
# qq_sum_path. A lot of this function is executed here in reversed order.
# This is because the i-th Gidney-adder needs to wait until the i+1-th finished.
# Therefore if we append the instructions of the i-th after the i+1-th, this 
# "waiting time" is realized.
def qq_sum_path_direct_uncomputation(a, b, c, R):
    
    if R == 1:
        # If R = 1, we are in the case of Drapers QCLA:
        # https://arxiv.org/abs/quant-ph/0406142   
        # We can use the formula
        # S = A (+) B (+) C
        for i in range(len(c)):
            cx(c[i], b[i+1])
        for i in range(len(a)):
            cx(a[i], b[i])
            
        # For this case, c will be uncomputed in qq_qcla.
        return
    
    # This loop treats the case that len(a) != len(b)
    # In this case we need to perform the increment function on the remaining
    # qubits of b (incase the carry is True)
    # #We start at the index the last loop finished at
    if len(a) != len(b):
        
        # Perform the loop in reverse
        for j in range(int(np.ceil(len(a)/R)), len(b)//R+2)[::-1]:
            
            b_block = b[R*j:R*(j+1)]
            
            if not len(b_block):
                continue
            
            #Perform incrementation function
            if R*j == 0:
                lin_incr(b_block, c_out = c[j])
            else:
                if j < len(c):
                    # We use the c_out of the incrementor to uncompute the carry of
                    # the previous iteration
                    lin_incr(b_block, c[j-1], c_out = c[j])
                else:
                    lin_incr(b_block, c[j-1])
                    

    # Execute addition using the corresponding carry values
    for i in range(len(a)//R+1)[::-1]:
        # Determine the radix qubits to perform the addition on
        a_block = a[R*i:R*(i+1)]
        b_block = b[R*i:R*(i+1)]
        
        # If a_block is empty we break the loop.
        # This doesn't necessarily imply that b_block is empty because
        # len(b) could be larger then len(a). In this case we execute an
        # increment block after this loop
        if not len(a_block):
            i = i-1
            continue
        
        
        # If a_block and b_block have different sizes, we create a padding
        # variable to fill up the remaining entries
        if not len(a_block) == len(b_block):
            padding_var = QuantumVariable(len(b_block) - len(a_block))
            a_block = a_block + list(padding_var)
        
        
        # Perform Gidney addition
        if i == 0:
            
            # Use the c_out keyword of the Gidney-Adder to uncompute the carry of
            # the previous iteration
            if len(c):
                gidney_adder(a_block, b_block, c_out = c[i])
            else:
                gidney_adder(a_block, b_block)
        
        
        elif i < len(c):
            gidney_adder(a_block, b_block, c_in = c[i-1], c_out = c[i])
        else:
            gidney_adder(a_block, b_block, c_in = c[i-1])
            
        
        # Delete carry var
        try:
            padding_var.delete(verify = bool(verify_manual_uncomputations[0]))
        except NameError:
            pass
    


"""
\********************************************************************************
* Copyright (c) 2023 the Qrisp authors
*
* This program and the accompanying materials are made available under the
* terms of the Eclipse Public License 2.0 which is available at
* http://www.eclipse.org/legal/epl-2.0.
*
* This Source Code may also be made available under the following Secondary
* Licenses when the conditions for such availability set forth in the Eclipse
* Public License, v. 2.0 are satisfied: GNU General Public License, version 2
* with the GNU Classpath Exception which is
* available at https://www.gnu.org/software/classpath/license.html.
*
* SPDX-License-Identifier: EPL-2.0 OR GPL-2.0 WITH Classpath-exception-2.0
********************************************************************************/
"""
from qrisp import *
from qrisp.qaoa import *
import scipy.stats as stats
import math
import networkx as nx

def test_TQA_warmstart():
    G = nx.Graph()
    edges = [(0, 5), (0, 3), (0, 7), (1, 3), (2, 4), (2, 7), (3, 4), (3, 7), (4, 6), (4, 7), (5, 7)]
    G.add_edges_from(edges)

    qarg = QuantumArray(qtype = QuantumVariable(1), shape = len(G))

    maxcut_instance = QAOAProblem(create_maxcut_cost_operator(G), RX_mixer, create_maxcut_cl_cost_function(G))

    count = 0
    repetitions=10
    for i in range (repetitions):
        print(i)
        benchmark_data1 = maxcut_instance.benchmark(qarg = QuantumVariable(len(G)),
                                depth_range = [3],
                                shot_range = [10000],
                                iter_range = [100],
                                optimal_solution = '00110110',
                                repetitions = 1
                                )

        #temp = benchmark_data1.rank(print_res = True)

        _,rndFO1=benchmark_data1.evaluate()

        approx1 = sum(rndFO1)/len(rndFO1)

        benchmark_data2 = maxcut_instance.benchmark(qarg = QuantumVariable(len(G)),
                                depth_range = [3],
                                shot_range = [10000],
                                iter_range = [100],
                                optimal_solution = '00110110',
                                init_type = 'tqa',
                                repetitions = 1
                                )

        #temp = benchmark_data2.rank(print_res = True)

        _,rndFO2=benchmark_data2.evaluate()

        approx2 = sum(rndFO2)/len(rndFO2)

        if approx1 > approx2:
            count += 1 

    # define the measured statistics of 1000 runs. p is the probability that the random start outperforms TQA warm start
    p = 489/1000

    # calculate p_empirical
    p_empirical = count / repetitions

    # calculate the test statistic
    z = (p_empirical - p) / math.sqrt((p * (1 - p)) / repetitions)

    # calculate the p-value
    p_value = 2 * (1 - stats.norm.cdf(abs(z)))

    # check if the p-value is less than the significance level
    assert p_value >= 0.01, "The test failed. The sample is unlikely to have come from the original probability p."


"""
\********************************************************************************
* Copyright (c) 2023 the Qrisp authors
*
* This program and the accompanying materials are made available under the
* terms of the Eclipse Public License 2.0 which is available at
* http://www.eclipse.org/legal/epl-2.0.
*
* This Source Code may also be made available under the following Secondary
* Licenses when the conditions for such availability set forth in the Eclipse
* Public License, v. 2.0 are satisfied: GNU General Public License, version 2
* with the GNU Classpath Exception which is
* available at https://www.gnu.org/software/classpath/license.html.
*
* SPDX-License-Identifier: EPL-2.0 OR GPL-2.0 WITH Classpath-exception-2.0
********************************************************************************/
"""

from qrisp import QFT, conjugate, p, cp, crz, cz, cx, x, QuantumBool, h, rzz, gphase, rz, QuantumCircuit, z, QuantumEnvironment, Operation, PGate
import numpy as np
from sympy import Symbol

def fourier_adder(a, b, perform_QFT = True):
    """
    In-place adder function based on `this paper <https://arxiv.org/abs/quant-ph/0410184>`_
    Performs the addition
    
    ::
        
        b += a
    

    Parameters
    ----------
    a : int or QuantumVariable or list[Qubit]
        The value that should be added.
    b : QuantumVariable or list[Qubit]
        The value that should be modified in the in-place addition.
    perform_QFT : bool, optional
        If set to ``False``, no QFT is executed. The default is ``True``.

    Examples
    --------
    
    We add two integers:
        
    >>> from qrisp import QuantumFloat, fourier_adder
    >>> a = QuantumFloat(4)
    >>> b = QuantumFloat(4)
    >>> a[:] = 4
    >>> b[:] = 5
    >>> fourier_adder(a,b)
    >>> print(b)
    {9: 1.0}

    """
    
    if perform_QFT:
        env = conjugate(QFT)(b, exec_swap = False)
    else:
        env = QuantumEnvironment()
    
    with env:
        
                    
        b = list(b)
        b = b[::-1]

        
        if isinstance(a, int):
            for i in range(len(b)):
                p(a*np.pi*2**(1+i-len(b)), b[i])
                
        else:
            
            if len(a) > len(b):
                raise Exception("Tried to add QuantumFloat of higher precision onto QuantumFloat of lower precision")
            
            phase_correction_a = np.zeros(len(a))
            phase_correction_b = np.zeros(len(b))
            for j in range(len(a)):
                for i in range(len(b)):    
                    
                    if 1+j+i-len(b) >= 1:
                        continue
                    if 1+j+i-len(b) == 0:
                        cz(a[j], b[i])
                    else:
                        b[i].qs().append(QuasiRZZ(-np.pi*2**(1+j+i-len(b))/2), [a[j], b[i]])
                        
                        phase_correction_a[j] += np.pi*2**(1+j+i-len(b))/2
                        phase_correction_b[i] += np.pi*2**(1+j+i-len(b))/2
            
            for i in range(len(b)):
                if phase_correction_b[i]%(2*np.pi) != 0:
                    p(phase_correction_b[i], b[i])
            for i in range(len(a)):
                if phase_correction_a[i]%(2*np.pi) != 0:
                    p(phase_correction_a[i], a[i])


class QuasiRZZ(Operation):
    
    def __init__(self, angle):
        qc = QuantumCircuit(2)
        qc.cx(qc.qubits[1], qc.qubits[0])
        qc.p(angle, qc.qubits[0])
        qc.cx(qc.qubits[1], qc.qubits[0])
        
        Operation.__init__(self, "quasi_rzz", num_qubits = 2, definition = qc, params = [angle])
        
        self.permeability = {0 : True, 1 : True}
        self.is_qfree = True        
    
    def inverse(self):
        return QuasiRZZ(-self.params[0])
    
    def control(self, num_ctrl_qubits=1, ctrl_state=-1, method=None):
        
        qc = QuantumCircuit(2 + num_ctrl_qubits)
        qc.cx(qc.qubits[-1], qc.qubits[-2])
        qc.append(PGate(self.params[0]).control(num_ctrl_qubits, ctrl_state=ctrl_state, method=method), qc.qubits[:-1])
        qc.cx(qc.qubits[-1], qc.qubits[-2])
        
        res = Operation.control(self, num_ctrl_qubits, ctrl_state=ctrl_state, method=method)
        
        res.definition = qc
        return res
        

# def helper_op(phi):
    
#     qc = helper_qc.bind_parameters({symb : phi})
#     res = qc.to_gate("quasi_rzz")
#     res.permeability = {0 : True, 1 : True}
#     res.is_qfree = True
#     return res


"""
\********************************************************************************
* Copyright (c) 2023 the Qrisp authors
*
* This program and the accompanying materials are made available under the
* terms of the Eclipse Public License 2.0 which is available at
* http://www.eclipse.org/legal/epl-2.0.
*
* This Source Code may also be made available under the following Secondary
* Licenses when the conditions for such availability set forth in the Eclipse
* Public License, v. 2.0 are satisfied: GNU General Public License, version 2
* with the GNU Classpath Exception which is
* available at https://www.gnu.org/software/classpath/license.html.
*
* SPDX-License-Identifier: EPL-2.0 OR GPL-2.0 WITH Classpath-exception-2.0
********************************************************************************/
"""

import time

import numpy as np
from scipy.optimize import minimize
from sympy import Symbol

from qrisp import h, QuantumArray, parallelize_qc
from qrisp.qaoa.qaoa_benchmark_data import QAOABenchmark


class QAOAProblem:
    r"""
    Central structure to facilitate treatment of QAOA problems.

    This class encapsulates the cost operator, mixer operator, and classical cost function for a specific QAOA problem instance. It also provides methods to set the initial state preparation function, classical cost post-processing function, and optimizer for the problem.

    For a quick demonstration, we import the relevant functions from already implemented problem instances:
        
    ::
        
        from networkx import Graph

        G = Graph()

        G.add_edges_from([[0,3],[0,4],[1,3],[1,4],[2,3],[2,4]])

        from qrisp.qaoa import (QAOAProblem, 
                                create_maxcut_cost_operator, 
                                create_maxcut_cl_cost_function,
                                RX_mixer)

        maxcut_instance = QAOAProblem(cost_operator = create_maxcut_cost_operator(G),
                                       mixer = RX_mixer,
                                       cl_cost_function = create_maxcut_cl_cost_function(G))
        
        
        from qrisp import QuantumVariable
        
        res = maxcut_instance.run(qarg = QuantumVariable(5),
                                  depth = 4, 
                                  max_iter = 25)
        
        print(res)
        #Yields: {'11100': 0.2847, '00011': 0.2847, '10000': 0.0219, '01000': 0.0219, '00100': 0.0219, '11011': 0.0219, '10111': 0.0219, '01111': 0.0219, '00010': 0.02, '11110': 0.02, '00001': 0.02, '11101': 0.02, '00000': 0.0173, '11111': 0.0173, '10010': 0.0143, '01010': 0.0143, '11010': 0.0143, '00110': 0.0143, '10110': 0.0143, '01110': 0.0143, '10001': 0.0143, '01001': 0.0143, '11001': 0.0143, '00101': 0.0143, '10101': 0.0143, '01101': 0.0143, '11000': 0.0021, '10100': 0.0021, '01100': 0.0021, '10011': 0.0021, '01011': 0.0021, '00111': 0.0021}
        
    For an in-depth tutorial, make sure to check out :ref:`MaxCutQAOA`!
        
    Parameters
    ----------
    cost_operator : function
        A function receiving a :ref:`QuantumVariable` or :ref:`QuantumArray` and parameter $\gamma$. This function should perform the application of the cost operator.
    mixer : function
        A function receiving a :ref:`QuantumVariable` or :ref:`QuantumArray` and parameter $\beta$. This function should perform the application mixing operator.
    cl_cost_function : function
        The classical cost function for the specific QAOA problem instance.

    """
    
    def __init__(self, cost_operator, mixer, cl_cost_function, init_function = None, init_type='random'):
        self.cost_operator = cost_operator
        self.mixer = mixer
        self.cl_cost_function = cl_cost_function
        
        self.init_function = init_function
        self.cl_post_processor = None
        self.init_type = init_type
        # Fourier heuristic parameterization
        self.fourier_depth = None
        # should be set in the 
        self.init_params = None

    def set_init_function(self, init_function):
        """
        Set the initial state preparation function for the QAOA problem.

        Parameters
        ----------
        init_function : function
            The initial state preparation function for the specific QAOA problem instance.
        """
        self.init_function = init_function
    
    def computeParams(self,p,t_max):
        """
        Compute the angle parameters gamma and beta based on the given inputs. Used for the TQA warm starting the initial state for QAOA.

        Parameters
        ----------
        p : int
            The number of partitions for the time interval.
        t_max : float
            The maximum time value.

        Returns
        -------
        np.array
            A concatenated numpy array of gamma and beta values.
        """
        dt = t_max / p
        t = dt * (np.arange(1, p + 1) - 0.5)
        gamma = (t / t_max) * dt
        beta = (1 - (t / t_max)) * dt
        return  np.concatenate((gamma,beta)) 

    def set_fourier_depth(self, fourier_depth, init_params = None):
        """
        Set the FOURIER heuristic for a QAOA problem.

        Parameters
        ----------
        fourier_depth : int
            Number of Fourier parameters.
        init_params : np.array
            (Optional) NumPy array of initial Fourier Parameters.
        """
        
        self.fourier_depth = fourier_depth
        self.init_params = init_params
        

      
    def compile_circuit(self, qarg, depth):
        """
        Compiles the circuit that is evaluated by the :meth:`run <qrisp.qaoa.QAOAProblem.run>` method.

        Parameters
        ----------
        qarg : QuantumVariable or QuantumArray
            The argument the cost function is called on.
        depth : int
            The amont of QAOA layers.


        Returns
        -------
        compiled_qc : QuantumCircuit
            The parametrized, compiled QuantumCircuit without measurements.
        list[sympy.Symbol]
            A list of the parameters that appear in ``compiled_qc``.

        Examples
        --------
        
        We create a MaxCut instance and compile the circuit:
        
            
        >>> from networkx import Graph
        >>> G = Graph([(0,1),(1,2),(2,0)])
        >>> from qrisp.qaoa import maxcut_problem
        >>> from qrisp import QuantumVariable
        >>> p = 5
        >>> qaoa_instance = maxcut_problem(G)
        >>> qarg = QuantumVariable(len(G))
        >>> qrisp_qc, symbols = qaoa_instance.compile_circuit(qarg, p)
        >>> print(qrisp_qc)
                     ┌───┐┌────────┐┌────────┐                          »
        qarg_dupl.0: ┤ H ├┤ gphase ├┤ gphase ├──■────────────────────■──»
                     ├───┤└────────┘└────────┘┌─┴─┐┌──────────────┐  │  »
        qarg_dupl.1: ┤ H ├────────────────────┤ X ├┤ P(2*gamma_0) ├──┼──»
                     ├───┤                    └───┘└──────────────┘┌─┴─┐»
        qarg_dupl.2: ┤ H ├─────────────────────────────────────────┤ X ├»
                     └───┘                                         └───┘»
        «                                            ┌──────────────┐   ┌────────┐   »
        «qarg_dupl.0: ───────■────────────────────■──┤ Rx(2*beta_0) ├───┤ gphase ├───»
        «                  ┌─┴─┐      ┌────────┐  │  └──────────────┘   └────────┘   »
        «qarg_dupl.1: ─────┤ X ├──────┤ gphase ├──┼─────────■────────────────────────»
        «             ┌────┴───┴─────┐└────────┘┌─┴─┐     ┌─┴─┐      ┌──────────────┐»
        «qarg_dupl.2: ┤ P(2*gamma_0) ├──────────┤ X ├─────┤ X ├──────┤ P(2*gamma_0) ├»
        «             └──────────────┘          └───┘     └───┘      └──────────────┘»
        «             ┌────────┐                                          »
        «qarg_dupl.0: ┤ gphase ├──────────────────■────────────────────■──»
        «             └────────┘┌──────────────┐┌─┴─┐┌──────────────┐  │  »
        «qarg_dupl.1: ────■─────┤ Rx(2*beta_0) ├┤ X ├┤ P(2*gamma_1) ├──┼──»
        «               ┌─┴─┐   ├──────────────┤└───┘└──────────────┘┌─┴─┐»
        «qarg_dupl.2: ──┤ X ├───┤ Rx(2*beta_0) ├─────────────────────┤ X ├»
        «               └───┘   └──────────────┘                     └───┘»
        «                                            ┌──────────────┐   ┌────────┐   »
        «qarg_dupl.0: ───────■────────────────────■──┤ Rx(2*beta_1) ├───┤ gphase ├───»
        «                  ┌─┴─┐      ┌────────┐  │  └──────────────┘   └────────┘   »
        «qarg_dupl.1: ─────┤ X ├──────┤ gphase ├──┼─────────■────────────────────────»
        «             ┌────┴───┴─────┐└────────┘┌─┴─┐     ┌─┴─┐      ┌──────────────┐»
        «qarg_dupl.2: ┤ P(2*gamma_1) ├──────────┤ X ├─────┤ X ├──────┤ P(2*gamma_1) ├»
        «             └──────────────┘          └───┘     └───┘      └──────────────┘»
        «             ┌────────┐                                          »
        «qarg_dupl.0: ┤ gphase ├──────────────────■────────────────────■──»
        «             └────────┘┌──────────────┐┌─┴─┐┌──────────────┐  │  »
        «qarg_dupl.1: ────■─────┤ Rx(2*beta_1) ├┤ X ├┤ P(2*gamma_2) ├──┼──»
        «               ┌─┴─┐   ├──────────────┤└───┘└──────────────┘┌─┴─┐»
        «qarg_dupl.2: ──┤ X ├───┤ Rx(2*beta_1) ├─────────────────────┤ X ├»
        «               └───┘   └──────────────┘                     └───┘»
        «                                            ┌──────────────┐   ┌────────┐   »
        «qarg_dupl.0: ───────■────────────────────■──┤ Rx(2*beta_2) ├───┤ gphase ├───»
        «                  ┌─┴─┐      ┌────────┐  │  └──────────────┘   └────────┘   »
        «qarg_dupl.1: ─────┤ X ├──────┤ gphase ├──┼─────────■────────────────────────»
        «             ┌────┴───┴─────┐└────────┘┌─┴─┐     ┌─┴─┐      ┌──────────────┐»
        «qarg_dupl.2: ┤ P(2*gamma_2) ├──────────┤ X ├─────┤ X ├──────┤ P(2*gamma_2) ├»
        «             └──────────────┘          └───┘     └───┘      └──────────────┘»
        «             ┌────────┐                                          »
        «qarg_dupl.0: ┤ gphase ├──────────────────■────────────────────■──»
        «             └────────┘┌──────────────┐┌─┴─┐┌──────────────┐  │  »
        «qarg_dupl.1: ────■─────┤ Rx(2*beta_2) ├┤ X ├┤ P(2*gamma_3) ├──┼──»
        «               ┌─┴─┐   ├──────────────┤└───┘└──────────────┘┌─┴─┐»
        «qarg_dupl.2: ──┤ X ├───┤ Rx(2*beta_2) ├─────────────────────┤ X ├»
        «               └───┘   └──────────────┘                     └───┘»
        «                                            ┌──────────────┐   ┌────────┐   »
        «qarg_dupl.0: ───────■────────────────────■──┤ Rx(2*beta_3) ├───┤ gphase ├───»
        «                  ┌─┴─┐      ┌────────┐  │  └──────────────┘   └────────┘   »
        «qarg_dupl.1: ─────┤ X ├──────┤ gphase ├──┼─────────■────────────────────────»
        «             ┌────┴───┴─────┐└────────┘┌─┴─┐     ┌─┴─┐      ┌──────────────┐»
        «qarg_dupl.2: ┤ P(2*gamma_3) ├──────────┤ X ├─────┤ X ├──────┤ P(2*gamma_3) ├»
        «             └──────────────┘          └───┘     └───┘      └──────────────┘»
        «             ┌────────┐                                          »
        «qarg_dupl.0: ┤ gphase ├──────────────────■────────────────────■──»
        «             └────────┘┌──────────────┐┌─┴─┐┌──────────────┐  │  »
        «qarg_dupl.1: ────■─────┤ Rx(2*beta_3) ├┤ X ├┤ P(2*gamma_4) ├──┼──»
        «               ┌─┴─┐   ├──────────────┤└───┘└──────────────┘┌─┴─┐»
        «qarg_dupl.2: ──┤ X ├───┤ Rx(2*beta_3) ├─────────────────────┤ X ├»
        «               └───┘   └──────────────┘                     └───┘»
        «                                            ┌──────────────┐                »
        «qarg_dupl.0: ───────■────────────────────■──┤ Rx(2*beta_4) ├────────────────»
        «                  ┌─┴─┐      ┌────────┐  │  └──────────────┘                »
        «qarg_dupl.1: ─────┤ X ├──────┤ gphase ├──┼─────────■────────────────────────»
        «             ┌────┴───┴─────┐└────────┘┌─┴─┐     ┌─┴─┐      ┌──────────────┐»
        «qarg_dupl.2: ┤ P(2*gamma_4) ├──────────┤ X ├─────┤ X ├──────┤ P(2*gamma_4) ├»
        «             └──────────────┘          └───┘     └───┘      └──────────────┘»
        «                                  
        «qarg_dupl.0: ─────────────────────
        «                  ┌──────────────┐
        «qarg_dupl.1: ──■──┤ Rx(2*beta_4) ├
        «             ┌─┴─┐├──────────────┤
        «qarg_dupl.2: ┤ X ├┤ Rx(2*beta_4) ├
        «             └───┘└──────────────┘
        """
        
        temp = list(qarg.qs.data)
        
        # Define QAOA angle parameters gamma and beta for QAOA circuit
        
        gamma = [Symbol("gamma_" + str(i)) for i in range(depth)]
        beta = [Symbol("beta_" + str(i)) for i in range(depth)]
        

        # Prepare the initial state for particular problem instance (MaxCut: superposition, GraphColoring: any node coloring)
        if self.init_function is not None:
            self.init_function(qarg)
        else:
            h(qarg)
            
        # Apply p layers of phase separators and mixers
        for i in range(depth):                           
            self.cost_operator(qarg, gamma[i])
            self.mixer(qarg, beta[i])

        # Compile quantum circuit with intended measurements    
        if isinstance(qarg, QuantumArray):
            intended_measurement_qubits = sum([list(qv) for qv in qarg.flatten()], [])
        else:
            intended_measurement_qubits = list(qarg)
        
        compiled_qc = qarg.qs.compile(intended_measurements=intended_measurement_qubits)
        
        qarg.qs.data = temp
        
        return compiled_qc, gamma + beta

    # Set initial random values for optimization parameters 
    #def optimization_routine(self, qarg, compiled_qc, symbols , depth, mes_kwargs, max_iter):    
    def optimization_routine(self, qarg, depth, mes_kwargs, max_iter): 
        """
        Wrapper subroutine for the optimization method used in QAOA. The initial values are set and the optimization via ``COBYLA`` is conducted here.

        Parameters
        ----------
        qarg : QuantumVariable or QuantumArray
            The argument the cost function is called on.
        complied_qc : QuantumCircuit
            The compiled quantum circuit.
        depth : int
            The amont of QAOA layers.
        symbols : list
            The list of symbols used in the quantum circuit.
        mes_kwargs : dict, optional
            The keyword arguments for the measurement function. Default is an empty dictionary, as defined in previous functions.
        max_iter : int, optional
            The maximum number of iterations for the optimization method. Default is 50, as defined in previous functions.
        init_type : string, optional
            Specifies the way the initial optimization parameters are chosen. Available are ``random`` and TQA. The default is ``random``.

        Returns
        -------
        res_sample
            The optimized parameters of the problem instance.
        """
        
        compiled_qc, symbols = self.compile_circuit(qarg, depth)
        
        # Set initial random values for optimization parameters 
    #    init_point = np.pi * np.random.rand(2 * depth)/2
        
        # initial point is set here, potentially subject to change
        if not isinstance(self.fourier_depth, int):
            init_point = np.pi * np.random.rand(2 * depth)/2
        elif not isinstance(self.init_params, list):
            init_point = np.pi * np.random.rand(2 * self.fourier_depth)/2
        else:

            if len(self.init_params)/2 != self.fourier_depth:
                raise Exception("Fourier-depth does not match match length of init_params! (should be half the length)")
            init_point = self.init_params

        # Define optimization wrapper function to be minimized using QAOA
        def optimization_wrapper(theta, qc, symbols, qarg, mes_kwargs):
            """
            Wrapper function for the optimization method used in QAOA.

            This function calculates the value of the classical cost function after post-processing if a post-processing function is set, otherwise it calculates the value of the classical cost function.

            Parameters
            ----------
            theta : list
                The list of angle parameters gamma and beta for the QAOA circuit.
            qc : QuantumCircuit
                The compiled quantum circuit.
            symbols : list
                The list of symbols used in the quantum circuit.
            qarg_dupl : QuantumVariable
                The duplicated quantum variable to which the quantum circuit is applied.
            mes_kwargs : dict
                The keyword arguments for the measurement function.

            Returns
            -------
            float
                The value of the classical cost function.
            """         
            subs_dic = {symbols[i] : theta[i] for i in range(len(symbols))}
            
            res_dic = qarg.get_measurement(subs_dic = subs_dic, precompiled_qc = qc, **mes_kwargs)
            
            cl_cost = self.cl_cost_function(res_dic)
            
            if self.cl_post_processor is not None:
                return self.cl_post_processor(cl_cost)
            else:
                return cl_cost
            
        def tqa_angles(p,qc, symbols, qarg_dupl, mes_kwargs,steps=10): #qarg only before
            """
            Compute the optimal parameters for the Time-dependent Quantum Adiabatic (TQA) algorithm.

            The function first creates a linspace array `time` from 0.1 to 10 with `steps` steps. 
            Then for each `t_max` in `time`, it computes the parameters `x` using the `computeParams` 
            function and calculates the energy `qcut` using the `optimization_wrapper` function. 
            The energy values are stored in the `energy` list. The `t_max` corresponding to the 
            minimum energy is found and used to compute the optimal parameters which are returned.

            Parameters
            ----------
            p : int
                The number of partitions for the time interval.
            qc : QuantumCircuit
                The quantum circuit for the specific problem instance.
            symbols : list
                The list of symbols in the quantum circuit.
            qarg_dupl : list
                The list of quantum arguments.
            mes_kwargs : dict
                The measurement keyword arguments.
            steps : int, optional
                The number of steps for the linspace function, default is 10.

            Returns
            -------
            np.array
                A concatenated numpy array of optimal gamma and beta values.
            """
            time = np.linspace(0.1, 10, steps)
            energy = []
            for t_max in time:      
                x=self.computeParams(p,t_max)
                qcut=optimization_wrapper(x,qc,symbols,qarg_dupl,mes_kwargs)
                energy.append(qcut)
            
            idx = np.argmin(energy)
            t_max = time[idx]
            return self.computeParams(p,t_max)

        if self.init_type=='random':
            # Set initial random values for optimization parameters
            init_point = np.pi * np.random.rand(2 * depth)/2

        elif self.init_type=='tqa':
            # TQA initialization
            init_point = tqa_angles(depth,compiled_qc, symbols, qarg, mes_kwargs)



        # Perform optimization using COBYLA method
        if isinstance(self.fourier_depth, int):
            from qrisp.qaoa.optimization_wrappers.fourier_wrapper import fourier_optimization_wrapper
            for index_p in range(1, depth + 1):
                compiled_qc, symbols = self.compile_circuit(qarg, index_p)
                res_sample =  minimize(fourier_optimization_wrapper,
                                    init_point, 
                                    method='COBYLA', 
                                    options={'maxiter':max_iter}, 
                                    args = (compiled_qc, symbols, qarg, self.cl_cost_function,self.fourier_depth, index_p , mes_kwargs))
                init_point = res_sample['x']
                
        else:
            compiled_qc, symbols = self.compile_circuit(qarg, depth)
            # Perform optimization using COBYLA method
            res_sample = minimize(optimization_wrapper,
                                init_point, 
                                method='COBYLA', 
                                options={'maxiter':max_iter}, 
                                args = (compiled_qc, symbols, qarg, mes_kwargs))
            
        return res_sample['x']
    
        

    def run(self, qarg, depth, mes_kwargs = {}, max_iter = 50, init_type = "random"):
        """
        Run the specific QAOA problem instance with given quantum arguments, depth of QAOA circuit,
        measurement keyword arguments (mes_kwargs) and maximum iterations for optimization (max_iter).
        
        Parameters
        ----------
        qarg : QuantumVariable
            The quantum variable to which the QAOA circuit is applied.
        depth : int
            The depth of the QAOA circuit.
        mes_kwargs : dict, optional
            The keyword arguments for the measurement function. Default is an empty dictionary.
        max_iter : int, optional
            The maximum number of iterations for the optimization method. Default is 50.
        init_type : string, optional
            Specifies the way the initial optimization parameters are chosen. Available are ``random`` and TQA. The default is ``random``.

        Returns
        -------
        opt_res : dict
            The optimal result after running QAOA problem for a specific problem instance. It contains the measurement results after applying the optimal QAOA circuit to the quantum variable.
        """

        #init_point = np.pi * np.random.rand(2 * depth)/2

        #alternative to everything below:
        #bound_qc = self.train_circuit(qarg, depth)
        #opt_res = bound_qc(qarg).get_measurement(**mes_kwargs)
        #return opt_res
        if not "shots" in mes_kwargs:
            mes_kwargs["shots"] = 5000
        
        #res_sample = self.optimization_routine(qarg, compiled_qc, symbols , depth,  mes_kwargs, max_iter)
        res_sample = self.optimization_routine(qarg, depth, mes_kwargs, max_iter)
        optimal_theta = res_sample 
        if isinstance(self.fourier_depth, int):
            from qrisp.qaoa.parameters.fourier_params import fourier_params_helper
            optimal_theta = fourier_params_helper(optimal_theta, self.fourier_depth , depth) 
        
        # Prepare initial state - in case this is not called, prepare superposition state
        if self.init_function is not None:
            self.init_function(qarg)
        else:
            h(qarg)

        # Apply p layers of phase separators and mixers    
        for i in range(depth):                          
            self.cost_operator(qarg, optimal_theta[i])
            self.mixer(qarg, optimal_theta[i+depth])
        opt_res = qarg.get_measurement(**mes_kwargs)
        
        return opt_res
    
    def train_function(self, qarg, depth, mes_kwargs = {}, max_iter = 50, init_type = "random"):
        """
        This function allows for training of a circuit with a given instance of a ``QAOAProblem``. It will then return a function that can be applied to a ``QuantumVariable``,
        s.t. that it is a solution to the problem instance. The function therefore acts as a circuit for the problem instance with optimized parameters.

        Parameters
        ----------
        qarg : QuantumVariable
            The quantum variable to which the QAOA circuit is applied.
        depth : int
            The depth of the QAOA circuit.
        mes_kwargs : dict, optional
            The keyword arguments for the measurement function. Default is an empty dictionary.
        max_iter : int, optional
            The maximum number of iterations for the optimization method. Default is 50.

        Returns
        -------
        circuit_generator : function
            A function that can be applied to a ```QuantumVariable`` , with optimized parameters for the problem instance. The ``QuantumVariable`` then represent a solution of the problem.

        Examples
        --------

        We create a MaxClique instance and train the ``QAOAProblem`` instance
        
        ::
            
            from qrisp.qaoa import QAOAProblem
            from qrisp.qaoa.problems.maxCliqueInfrastr import maxCliqueCostfct,maxCliqueCostOp,init_state
            from qrisp.qaoa.mixers import RX_mixer
            from qrisp import QuantumVariable
            import networkx as nx
            
	        #create QAOAinstance
            G = nx.erdos_renyi_graph(9,0.7, seed =  133)
	        QAOAinstance = QAOAProblem(maxCliqueCostOp(G), RX_mixer, maxCliqueCostfct(G))
	        QAOAinstance.set_init_function(init_function=init_state)

            # create a blueprint-qv to train the problem instance on and train it
            qarg_new = QuantumVariable(G.number_of_nodes())
            training_func = QAOAinstance.train_circuit( qarg=qarg_new, depth=5 )

            # apply the trained function to a new qv 
            qarg_trained = QuantumVariable(G.number_of_nodes())
            training_func(qarg_trained)

            # get the results in a nice format
            opt_res = qarg_trained.get_measurement()
            aClCostFct = maxCliqueCostfct(G)

            print("5 most likely Solutions") 
            maxfive = sorted(opt_res, key=opt_res.get, reverse=True)[:5]
            for res, val in opt_res.items():  
                if res in maxfive:

                    print((res, val))
                    print(aClCostFct({res : 1})) 

        """

        compiled_qc, symbols = self.compile_circuit(qarg, depth)
        res_sample = self.optimization_routine(qarg, depth, mes_kwargs, max_iter)
        
        def circuit_generator(qarg_gen):
            if self.init_function is not None:
                self.init_function(qarg_gen)
            else:
                h(qarg_gen)
            for i in range(depth): 
                
                self.cost_operator(qarg_gen, res_sample[i])
                self.mixer(qarg_gen, res_sample[i+depth])
            
        return circuit_generator
    
    def benchmark(self, qarg, depth_range, shot_range, iter_range, optimal_solution, repetitions = 1, mes_kwargs = {}, init_type = "random"):
        """
        This method enables convenient data collection regarding performance of the implementation.

        Parameters
        ----------
        qarg : QuantumVariable or QuantumArray
            The quantum argument, the benchmark is executed on. Compare to the :meth:`.run <qrisp.qaoa.QAOAProblem.run>` method.
        depth_range : list[int]
            A list of integers indicating, which depth parameters should be explored. Depth means the amount of QAOA layers.
        shot_range : list[int]
            A list of integers indicating, which shots parameters should be explored. Shots means the amount of repetitions, the backend performs per iteration.
        iter_range : list[int]
            A list of integers indicating, what iterations parameter should be explored. Iterations means the amount of backend calls, the optimizer is allowed to do.
        optimal_solution : -
            The optimal solution to the problem. Should have the same type as the keys of the result of ``qarg.get_measurement()``.
        repetitions : int, optional
            The amount of repetitions, each parameter constellation should go though. Can be used to get a better statistical significance. The default is 1.
        mes_kwargs : dict, optional
            The keyword arguments, that are used for the ``qarg.get_measurement``. The default is {}.
        init_type : string, optional
            Specifies the way the initial optimization parameters are chosen. Available are ``random`` and TQA. The default is ``random``.

        Returns
        -------
        :ref:`QAOABenchmark`
            The results of the benchmark.
            
        Examples
        --------
        
        We create a MaxCut instance and benchmark several parameters
        
        ::
            
            from qrisp import *
            from networkx import Graph
            G = Graph()

            G.add_edges_from([[0,3],[0,4],[1,3],[1,4],[2,3],[2,4]])

            from qrisp.qaoa import maxcut_problem

            max_cut_instance = maxcut_problem(G)

            benchmark_data = max_cut_instance.benchmark(qarg = QuantumVariable(5),
                                       depth_range = [3,4,5],
                                       shot_range = [5000, 10000],
                                       iter_range = [25, 50],
                                       optimal_solution = "11100",
                                       repetitions = 2
                                       )
        
        We can investigate the data by calling ``visualize``:
        
        ::
            
            benchmark_data.visualize()
        
        .. image:: benchmark_plot.png
            
        The :ref:`QAOABenchmark` class contains a variety of methods to help 
        you drawing conclusions from the collected data. Make sure to check them out!

        """
        
        data_dict = {"layer_depth" : [],
                     "circuit_depth" : [],
                     "qubit_amount" : [],
                     "shots" : [],
                     "iterations" : [],
                     "counts" : [],
                     "runtime" : [],
                     "cl_cost" : []
                     }
        
        for p in depth_range:
            for s in shot_range:
                for it in iter_range:
                    for k in range(repetitions):
                        
                        if isinstance(qarg, QuantumArray):
                            qarg_dupl = QuantumArray(qtype = qarg.qtype, shape = qarg.shape)
                            mes_qubits = sum([qv.reg for qv in qarg_dupl.flatten()], [])
                        else:
                            qarg_dupl = qarg.duplicate()
                            mes_qubits = list(qarg_dupl)
                            
                        start_time = time.time()
                        
                        temp_mes_kwargs = dict(mes_kwargs)
                        temp_mes_kwargs["shots"] = s
                        if init_type=='random':
                            counts = self.run(qarg=qarg_dupl, depth = p, max_iter = it, mes_kwargs = temp_mes_kwargs, init_type='random')
                        elif init_type=='tqa':
                            counts = self.run(qarg=qarg_dupl, depth = p, max_iter = it, mes_kwargs = temp_mes_kwargs, init_type='tqa')
                        final_time = time.time() - start_time
                        
                        compiled_qc = qarg_dupl.qs.compile(intended_measurements=mes_qubits)
                        
                        data_dict["layer_depth"].append(p)
                        data_dict["circuit_depth"].append(compiled_qc.depth())
                        data_dict["qubit_amount"].append(compiled_qc.num_qubits())
                        data_dict["shots"].append(s)
                        data_dict["iterations"].append(it)
                        data_dict["counts"].append(counts)
                        data_dict["runtime"].append(final_time)
                        
                        
        return QAOABenchmark(data_dict, optimal_solution, self.cl_cost_function)
    

    def visualize_cost(self):
        """
        TODO

        Returns
        -------
        None.

        """
        pass



"""
\********************************************************************************
* Copyright (c) 2023 the Qrisp authors
*
* This program and the accompanying materials are made available under the
* terms of the Eclipse Public License 2.0 which is available at
* http://www.eclipse.org/legal/epl-2.0.
*
* This Source Code may also be made available under the following Secondary
* Licenses when the conditions for such availability set forth in the Eclipse
* Public License, v. 2.0 are satisfied: GNU General Public License, version 2
* with the GNU Classpath Exception which is
* available at https://www.gnu.org/software/classpath/license.html.
*
* SPDX-License-Identifier: EPL-2.0 OR GPL-2.0 WITH Classpath-exception-2.0
********************************************************************************/
"""


from qrisp.interface import BackendClient

api_endpoint = "127.0.0.1"

def CirqSim():
    return BackendClient(api_endpoint, port = 8083)

def PennylaneSim():
    return BackendClient(api_endpoint, port = 8084)
    
def MQTSim():
    return BackendClient(api_endpoint, port = 8085)

def PennylaneRigettiSim():
    return BackendClient(api_endpoint, port = 8086)

def PyTketStimSim():
    return BackendClient(api_endpoint, port = 8087)

def QulacsSim():
    return BackendClient(api_endpoint, port = 8088)

def QSimCirq():
    return BackendClient(api_endpoint, port = 8089)
    
def QiboSim():
    return BackendClient(api_endpoint, port = 8090)

def QrispSim():
    return BackendClient(api_endpoint, port = 8090)

def QiskitSim():
    return BackendClient(api_endpoint, port = 8090)    


from qrisp import rz, rzz, x
import numpy as np
import copy
from qrisp.qiro.qiroproblems.qiro_utils import * 

def create_maxIndep_replacement_routine( res, Graph, solutions= [], exclusions= []):
    """
    Creates a replacement routine for the problem structure, i.e. defines the replacement rules. 
    See the original paper for description of the update rules

    Parameters
    ----------
    res : dict
        Result dictionary of initial QAOA optimization procedure.
    Graph : nx.Graph
        The Graph defining the problem instance.
    solutions : List
        Qubits which have been found to be positive correlated, i.e. part of the problem solution.
    exclusions : List
        Qubits which have been found to be negative correlated, i.e. not part of the problem solution, or contradict solution qubits in accordance to the update rules.  

    Returns
    -------
    newGraph : nx.Graph
        Updated graph of the problem instance.
    solutions : List
        Updated set of solutions to the problem.
    sign : Int
        The sign of the correlation.
    exclusions : List
        Updated set of exclusions to the problem.



        
        
    """
    # For multi qubit correlations
    orig_edges = [list(item) for item in Graph.edges()]

    # FOR SINGLE QUBIT CORRELATIONS
    orig_nodes = list(Graph.nodes())
    
    max_item = []
    max_item, sign = find_max(orig_nodes, orig_edges , res, solutions)
    newGraph = copy.deepcopy(Graph)

    # we just directly remove vertices from the graph 
    if isinstance(max_item, int):
        if sign > 0:
            to_remove = Graph.adj[max_item]
            newGraph.remove_nodes_from(to_remove)
            solutions.append(max_item)
            exclusions += to_remove

        elif sign < 0:
            newGraph.remove_node(max_item)
            exclusions.append(max_item)

    else:
        if sign > 0:
            newGraph.remove_nodes_from(max_item)
            exclusions += list(max_item)

        elif sign < 0:
            #remove 
            intersect = list(set( list(Graph.adj[max_item[0]].keys()) ) & set( list(Graph.adj[max_item[0]].keys()) ))
            newGraph.remove_nodes_from(intersect)
            exclusions += intersect 

    return newGraph, solutions, sign, exclusions


def create_maxIndep_cost_operator_reduced(Graph, solutions= []):
    """
    |  Based on PennyLane unconstrained mixer implementation.
    |  Initial state in :math:`(|0>+|1>)^{\otimes n}` . 
    |  This operator is then adjusted to consider qubits that have been found to be a part of the problem solution.

    Parameters
    ----------
    G : nx.Graph
        Graph of the problem instance
    solutions : List
        Qubits which have been found to be positive correlated, i.e. part of the problem solution.
    
    Returns
    -------
    partialCostMixer : function
        The Operator to be applied to the problem ``QuantumVariable``

    """
    def partialcostMixer(qv, gamma):
        for pair in list(Graph.edges()):
            #cx(qv[pair[0]], qv[pair[1]])
            rzz(3*gamma, qv[pair[0]], qv[pair[1]])
            rz(-gamma, qv[pair[0]])
            rz(-gamma, qv[pair[1]])
        for i in Graph.nodes():
            if not i in solutions:
                rz(gamma, qv[i])
        #return qv

    return partialcostMixer



""" def create_maxIndep_mixer_reduced(Graph, solutions):

    def RX_mixer(qv, beta):

        from qrisp import rx
        for i in Graph.nodes():
            if not i in solutions:
                rx(2 * beta, qv[i])
    return RX_mixer


def init_function_reduced(Graph, solutions):

    def init_state(qv):
        from qrisp import h
        for i in Graph.nodes():
            if not i in solutions:
                h(qv[i])
        for i in solutions:
            x(qv[i])
    return init_state



#TODO:
def create_maxIndep_cl_cost_function_reduced(Graph):

        #btw alternative formulation: for edge: check if string[edge[0]] != string[edge[1]] 
    def aClcostFct(res_dic):
        tot_energy = 0.001
        tot_counts = 0
        for state in res_dic.keys():
            # we assume solution is right
            temp = True
            energy = 0 
            for edge in Graph.edges():
                if not state[edge[0]] != state[edge[1]]:
                    temp = False 
            
            # else we just add the number of marked as |1> nodes
            if temp: 
                intlist = [s for s in range(len(list(state))) if list(state)[s] == "1"]
                energy = -len(intlist)
            
            tot_energy += energy * res_dic[state]
            tot_counts += res_dic[state]

        #print(tot_energy/tot_counts)

        return tot_energy/tot_counts

    return aClcostFct 
"""


"""
\********************************************************************************
* Copyright (c) 2023 the Qrisp authors
*
* This program and the accompanying materials are made available under the
* terms of the Eclipse Public License 2.0 which is available at
* http://www.eclipse.org/legal/epl-2.0.
*
* This Source Code may also be made available under the following Secondary
* Licenses when the conditions for such availability set forth in the Eclipse
* Public License, v. 2.0 are satisfied: GNU General Public License, version 2
* with the GNU Classpath Exception which is
* available at https://www.gnu.org/software/classpath/license.html.
*
* SPDX-License-Identifier: EPL-2.0 OR GPL-2.0 WITH Classpath-exception-2.0
********************************************************************************/
"""

from qrisp import QFT, conjugate, p, cp, QuantumFloat, QuantumVariable
import numpy as np


def ammend_inpl_adder(raw_inpl_adder, ammend_cl_int = True):
    
    
    def ammended_adder(qf2, qf1, *args, ignore_rounding_error = False, ignore_overflow_error = True, **kwargs):
    
        
    
        if isinstance(qf1, QuantumFloat) and isinstance(qf2, QuantumFloat) and False:
            qs = qf1.qs
        
            # If qf2 has lower exponent, the qf2 bits with less significance than all of qf1
            # can not be added to qf1 (rounding error)
            if not ignore_rounding_error and qf1.exponent > qf2.exponent:
                raise Exception(
                    "Tried to add QuantumFloat to QuantumFloat of lower precision"
                    " (set ignore_rounding_error = True)"
                )
        
            # If qf2 has higher maximum significance than qf1, the qf2 bits with higher
            # significance than all of qf1 can not be added to qf1 (overflow error)
            if not ignore_overflow_error and qf1.mshape[1] < qf2.mshape[1]:
                raise Exception(
                    "Tried to add QuantumFloat to QuantumFloat of lower precision"
                    " (set ignore_overflow_error = True)"
                )
        
            # Determine the significance range
            # For instance for a QuantumFloat with datashape [-2,3]
            # has significance range [-2,-1,0,1,2,3]
            significance_range_qf1 = list(range(qf1.mshape[0], qf1.mshape[1] + 1))
            significance_range_qf2 = list(range(qf2.mshape[0], qf2.mshape[1] + 1))
        
            # Determine the intersection of the significance ranges
            signficance_range_intersetion = list(
                set(significance_range_qf1).intersection(significance_range_qf2)
            )
        
            # Determine maximum and minimum significance of the addition
            # The maximum significance is the maximum significance of qf1
            # (since we are adding in-place)
            # The minimum significance is the minimum of the intersection because
            # if qf2.exponent < qf1.exponent
            # we are allowed to ignore possible rounding errors
            # (skip all significances lower than qf1.exponent)
            # if qf2.exponent > qf1.exponent
            # The qbits of qf1 with lower significance than all of qf2
            # stay unchanged during addition
        
            max_sig = max(significance_range_qf1)
            min_sig = min(signficance_range_intersetion)
        
            # If the maximum significance is higher than the maximum signficance of qf2, we need
            # to augment some ancilla qubits because the Cuccaro-procedure requires equal
            # amount of input qubits
        
        
            if max_sig > max(significance_range_qf2):
                # print(max_sig-max(significance_range_qf2))
                ancilla_var = QuantumVariable(
                    max_sig - max(significance_range_qf2) - int(qf2.signed)
                )
                augmented_qf2_qbs = qf2.reg + ancilla_var.reg
            else:
                augmented_qf2_qbs = qf2.reg
        
            # Determine the bit window of which bits should participate in the Cuccaro-procedure
            bit_window_1 = [
                significance_range_qf1.index(min_sig),
                significance_range_qf1.index(max_sig),
            ]
        
            if max_sig > max(significance_range_qf2):
                bit_window_2 = [significance_range_qf2.index(min_sig), len(augmented_qf2_qbs)]
            else:
                bit_window_2 = [
                    significance_range_qf2.index(min_sig),
                    significance_range_qf2.index(max_sig),
                ]
        
            # print(max_sig)
            # print(ancilla_var_.size)
            # print(significance_range_qf1)
            # print(significance_range_qf2)
            # print(bit_window_1)
            # print(bit_window_2)
        
            # Determine qubit lists
            qubit_list_1 = qf1.reg[bit_window_1[0] : bit_window_1[1]]
            qubit_list_2 = augmented_qf2_qbs[bit_window_2[0] : bit_window_2[1]]
        
            # print(len(qubit_list_1))
            # print(len(qubit_list_2))
        
            # Now we treat the sign bits
            if qf2.signed and not qf1.signed:
                raise Exception("Tried to add signed QuantumFloat to non signed QuantumFloat")
        
            # The signs are handled via modular arithmetic
            # According to the Cuccaro-Paper this is done via manipulating
            # the final qubits of the input list
            # if True:
            if qf1.signed:
                qubit_list_1.append(qf1[-1])
        
                if qf2.signed:
                    qubit_list_2.append(qf2[-1])
                else:
                    ancilla_var_2 = QuantumVariable(1)
                    qubit_list_2.append(ancilla_var_2[0])
        
            # If qf2 is signed and we augmented Qubits, we need to make sure
            # the augmented qubits are set to 1 if qf2 is in a negative state
            # (for details on this check how negative numbers are encoded)
            if qf2.signed and max_sig > max(significance_range_qf2):
                for i in range(ancilla_var.size):
                    qs.cx(qf2[-1], ancilla_var[i])
        
            if len(qubit_list_1) > 1:
                raw_inpl_adder(qubit_list_1[:-1], qubit_list_2[:-1], *args, carry_out = qubit_list_1[-1], **kwargs)
        
            # Perform 2nd step of the modular addition section
            # qs.cx(qubit_list_2[-2], qubit_list_1[-1])
            qs.cx(qubit_list_2[-1], qubit_list_1[-1])
        
            # Remove the 1s from the augmented qubits and delete
            if len(augmented_qf2_qbs) != len(qf2.reg):
                if qf2.signed:
                    for i in range(ancilla_var.size):
                        qs.cx(qf2[-1], ancilla_var[i])
        
            try:
                ancilla_var.delete()
            except NameError:
                pass
        
            if qf1.signed and not qf2.signed:
                ancilla_var_2.delete()
                
        elif isinstance(qf2, (list, QuantumVariable)):
            
            if len(qf2) < len(qf1):
                ancilla_var = QuantumVariable(len(qf1)-len(qf2), name = "add_ammend_anc*", qs = qf1[0].qs())
                qf2 = list(qf2) + list(ancilla_var)
            
            raw_inpl_adder(qf2, qf1, *args, **kwargs)
            
            try:
                ancilla_var.delete(verify = False)
            except NameError:
                pass
            
        elif isinstance(qf2, int) and ammend_cl_int:
            
            ancilla_var = QuantumFloat(len(qf1), name = "add_ammend_anc*", qs = qf1[0].qs())
            
            ancilla_var.encode(qf2 % 2**len(qf1))
            
            raw_inpl_adder(ancilla_var, qf1, *args, **kwargs)
            ancilla_var.encode(qf2 % 2**len(qf1), permit_dirtyness=True)
            
            ancilla_var.delete(verify = False)
        
        else:
            raw_inpl_adder(qf2, qf1, *args, **kwargs)
            
    
    return ammended_adder


"""
\********************************************************************************
* Copyright (c) 2023 the Qrisp authors
*
* This program and the accompanying materials are made available under the
* terms of the Eclipse Public License 2.0 which is available at
* http://www.eclipse.org/legal/epl-2.0.
*
* This Source Code may also be made available under the following Secondary
* Licenses when the conditions for such availability set forth in the Eclipse
* Public License, v. 2.0 are satisfied: GNU General Public License, version 2
* with the GNU Classpath Exception which is
* available at https://www.gnu.org/software/classpath/license.html.
*
* SPDX-License-Identifier: EPL-2.0 OR GPL-2.0 WITH Classpath-exception-2.0
********************************************************************************/
"""

from qrisp import *
import numpy as np
from operator import itemgetter

def QUBO_obj(bitstring, Q):
    x = np.array(list(bitstring), dtype=int)
    cost = x.T @ Q @ x
    return cost

def create_QUBO_cl_cost_function(Q):
    """
    Creates the classical cost function for QUBO with the QUBO matrix `Q`` that we are attempting to solve.

    Parameters
    ----------
    Q : np.array
        QUBO matrix to solve.

    Returns
    -------
    cl_cost_function : function
        Classical cost function, which in the end returns the ratio between 
        the energy calculated using the QUBO_obj objective function and the 
        amount of counts used in the experiment.

    """    
    def cl_cost_function(counts):
        
        def QUBO_obj(bitstring, Q):
            x = np.array(list(bitstring), dtype=int)
            cost = x.T @ Q @ x
            return cost
        
        energy = 0
        for meas, meas_count in counts.items():
            obj_for_meas = QUBO_obj(meas,Q)
            energy += obj_for_meas * meas_count
        return energy
    
    return cl_cost_function


# MAPPING QUBO TO GATES                       1 - Z_i                 1
#                               x_i  <-----> ---------:     x^T Q x = - SUM_{i,j}  Q_{ij} (1-Z_i)(1-Z_j) = SUM_{i,j} Q_ij*(Z_i*Z_j - Z_i - Z_j + 1)/4
#                                                2                    4  


def create_QUBO_cost_operator(Q):
    """
    Creates the QUBO operator as a sequence of unitary gates. In the QAOA overview section this is also called the phase separator $U_P$.

    Parameters
    ----------
    Q : np.array
        QUBO matrix to solve.

    Returns
    -------
    cost_operator function.

    """

    def QUBO_cost_operator(qv, gamma):

        gphase(-gamma/4*(np.sum(Q)+np.trace(Q)),qv[0])

        for i in range(len(Q)):
            rz(-gamma/2*(sum(Q[i])+sum(Q[:,i])), qv[i])
            for j in range(len(Q)):
                if i != j and Q[i][j] != 0:
                    rzz(gamma/2*Q[i][j], qv[i], qv[j])
    return QUBO_cost_operator


def QUBO_problem(Q):
    """
    Creates a QAOA problem instance taking the phase separator, appropriate mixer, and
    appropriate classical cost function into account.

    Parameters
    ----------
    Q : np.array
        QUBO matrix to solve.

    Returns
    -------
    QAOAProblem : function
        QAOA problem instance for QUBO with which the QAOA algorithm is ran for.

    """    
    from qrisp.qaoa import QAOAProblem, RX_mixer
    
    return QAOAProblem(create_QUBO_cost_operator(Q), RX_mixer, create_QUBO_cl_cost_function(Q))


def solve_QUBO(Q, depth, max_iter = 50, backend = None, n_solutions = 1, print_res = False):
    """
    Solves a Quadratic Unconstrained Binary Optimization (QUBO) problem using the Quantum Approximate Optimization Algorithm (QAOA). 
    The function imports the default backend from the 'qrisp.default_backend' module. 
    It defines a quantum argument as a QuantumArray of len(Q) QuantumVariables with size 1. 
    It then runs the QAOA with the given quantum arguments, ``depth``, measurement keyword arguments, and a maximum of 50 iterations for optimization. 
    The functions then considers the first ``n_solutions`` solutions of the QAOA optimization, and as a classical post processing, 
    calculates the cost for each such solution, sorts the solutions by their cost in ascending order, and prints the solutions with their corresponding costs.

    Parameters
    ----------
    Q : np.array
        QUBO matrix to solve.
    depth : int
        The depth (amount of layers) of the QAOA circuit.
    max_iter : int
        The maximal amount of iterations of the COBYLA optimizer in the QAOA algorithm.
    backend : str
        The backend to be used for the quantum/annealing simulation.
    n_solutions : int
        The number of solutions to consider for classical post processing. The defalut is 1.

    Returns
    -------
    optimal_solution: tuple
        The function returns the optimal solution as a tuple where the first element is the cost and the second element is the optimal bitstring. If print_res is set to True, the function prints the runtime of the QAOA algorithm and the ``n_solutions`` best solutions with their respective costs.

    """

    # Define quantum argument as a QuantumArray of len(G) QuantumVariables with size 1 or as a QuantumVariable with size len(G)
    qarg = QuantumArray(qtype = QuantumVariable(1), shape = len(Q))

    QUBO_instance = QUBO_problem(Q)

    if backend is None:
        from qrisp.default_backend import def_backend
        backend = def_backend

    else:
        backend = backend

    # Run QAOA with given quantum arguments, depth, measurement keyword arguments and maximum iterations for optimization
    res = QUBO_instance.run(qarg, depth, mes_kwargs={"backend" : backend}, max_iter = max_iter) # runs the simulation
    res = dict(list(res.items())[:n_solutions])

    # Calculate the cost for each solution
    costs_and_solutions = [(QUBO_obj(bitstring, Q), bitstring) for bitstring in res.keys()]

    # Sort the solutions by their cost in ascending order
    sorted_costs_and_solutions = sorted(costs_and_solutions, key=itemgetter(0))

    optimal_solution = sorted_costs_and_solutions[0]

    if print_res is True:
        # Get the top solutions and print them
        for i in range(n_solutions):
            print(f"Solution {i+1}: {sorted_costs_and_solutions[i][1]} with cost: {sorted_costs_and_solutions[i][0]}")
    
    return optimal_solution




"""
\********************************************************************************
* Copyright (c) 2023 the Qrisp authors
*
* This program and the accompanying materials are made available under the
* terms of the Eclipse Public License 2.0 which is available at
* http://www.eclipse.org/legal/epl-2.0.
*
* This Source Code may also be made available under the following Secondary
* Licenses when the conditions for such availability set forth in the Eclipse
* Public License, v. 2.0 are satisfied: GNU General Public License, version 2
* with the GNU Classpath Exception which is
* available at https://www.gnu.org/software/classpath/license.html.
*
* SPDX-License-Identifier: EPL-2.0 OR GPL-2.0 WITH Classpath-exception-2.0
********************************************************************************/
"""

from qrisp import h, rz , cx
import numpy as np
from scipy.optimize import minimize
from sympy import Symbol
import itertools

from collections.abc import Iterable 

# rethink this --> can we call clause[3]% 2 in assignment? because otherwise his blows up the circuit size...


def eThrTwocostOp(clauses):

    """
    |  Implementation for eThrTwoLin Cost-Operator.
    |  Straight forward, apply :math:`R_{ZZZ}` gates between all qubits meantioned in each clause.

    
    Parameters
    ----------
    clauses : List(List)
        clauses to be considered for eTwoThrLin
        
    Returns
    -------
    QuantumCircuit: qrisp.QuantumCircuit
        the Operator applied to the circuit-QuantumVariable

    Examples
    --------

    first three are the indices (a_1,j ; a_2,j ; a3,j), last one is b_L
    
    >>> clauses = [(1,3,4,5), (2,6,9,11), (1,5,8,33), (3,4,5,77), (2,4,6,7), (1,3,4,55), (1,3,6,49), (1,3,4,57), (1,3,6,44), (2,4,7,8)]
    >>> cost_operator = eThrTwocostOp(clauses=clauses)

    """
    if not isinstance(clauses, Iterable):
        raise Exception("Wrong structure of problem - clauses have to be iterable!")
    for clause in clauses:
        if not isinstance(clause, Iterable):
            raise Exception("Wrong structure of problem - each clauses has to be an array!")
        for item in clause:
            if not isinstance(item, int):
                raise Exception("Wrong structure of problem - each literal has to an int!")


    def eThrTwocostnoEmbed(qv , gamma):
        #clauses = clauses
        qc = qv
        for numClause in range(len(clauses)):
            #go through clauses and apply quantum gates 
            clause = clauses[numClause]

            cx(qc[clause[0]], qc[clause[1]])
            cx(qc[clause[1]], qc[clause[2]])
            rz(pow(-1, clause[3] % 2)* gamma, qc[clause[2]])
            cx(qc[clause[1]], qc[clause[2]])
            cx(qc[clause[0]], qc[clause[1]])

        return qc
    
    return eThrTwocostnoEmbed



def eTwoThrLinclCostfct(clauses):
    
    """

    Parameters
    ----------
    clauses : List
            clauses to analyze in the problem instance

    Returns
    -------
    Costfunction : function
        the classical function for the problem instance, which takes a dictionary of measurement results as input
    """

    def setupaClCostfct(res_dic):
        energy = 0
        total_counts = 0
        for state in list(res_dic.keys()):
            obj = 0
            #ljust do a minus 1 op if state is equiv to a given condition clause
            for index in range(len(clauses)):
                clause = clauses[index]
                sum = 0
                for index_aj in range(0,2):
                    sum += int(state[clause[index_aj]])
                if sum == clause[3] % 2:
                    obj -= 1 
            energy += obj * res_dic[state]
            total_counts += res_dic[state]
        #print(energy/total_counts)

        return energy/total_counts
    
    return setupaClCostfct


def init_state(qv):
    # all in 0
    h(qv)
    return qv




"""
\********************************************************************************
* Copyright (c) 2023 the Qrisp authors
*
* This program and the accompanying materials are made available under the
* terms of the Eclipse Public License 2.0 which is available at
* http://www.eclipse.org/legal/epl-2.0.
*
* This Source Code may also be made available under the following Secondary
* Licenses when the conditions for such availability set forth in the Eclipse
* Public License, v. 2.0 are satisfied: GNU General Public License, version 2
* with the GNU Classpath Exception which is
* available at https://www.gnu.org/software/classpath/license.html.
*
* SPDX-License-Identifier: EPL-2.0 OR GPL-2.0 WITH Classpath-exception-2.0
********************************************************************************/
"""


from qrisp.interface import VirtualBackend, VirtualQiskitBackend
# from qrisp.interface.qunicorn import BackendClient
# def_backend = VirtualQiskitBackend()
from qrisp.simulator.simulator import run
from qrisp import QuantumCircuit
#

def run_wrapper(qasm_str, shots, token = ""):
    qc = QuantumCircuit.from_qasm_str(qasm_str)
    return run(qc, shots)

# def_backend = VirtualBackend(run_wrapper, port=8080)

class DefaultBackend:
    
    def run(self, qc, shots, token = ""):
        
        return run(qc, shots, token)


def_backend = DefaultBackend()
# def_backend = BackendClient()
# raise


"""
\********************************************************************************
* Copyright (c) 2023 the Qrisp authors
*
* This program and the accompanying materials are made available under the
* terms of the Eclipse Public License 2.0 which is available at
* http://www.eclipse.org/legal/epl-2.0.
*
* This Source Code may also be made available under the following Secondary
* Licenses when the conditions for such availability set forth in the Eclipse
* Public License, v. 2.0 are satisfied: GNU General Public License, version 2
* with the GNU Classpath Exception which is
* available at https://www.gnu.org/software/classpath/license.html.
*
* SPDX-License-Identifier: EPL-2.0 OR GPL-2.0 WITH Classpath-exception-2.0
********************************************************************************/
"""

from qrisp.qaoa import QAOAProblem
from qrisp.qaoa.problems.maxKColorableSubgraph import apply_phase_if_eq,create_coloring_operator,create_coloring_cl_cost_function,QuantumColor
from qrisp.qaoa import XY_mixer, apply_XY_mixer, RX_mixer
from qrisp import QuantumArray
import random
import networkx as nx
from operator import itemgetter


from qrisp.default_backend import def_backend
qrisp_sim = def_backend
qaoa_backend = qrisp_sim

depth = 3

def mkcs_obj(quantumcolor_array, G):
        # Set value of color integer to 1
    color = 1

        # Iterate over all edges in graph G
    for pair in list(G.edges()):

            # If colors of nodes in current pair are not same, multiply color by reward factor 4
        if quantumcolor_array[pair[0]] != quantumcolor_array[pair[1]]:
            color *= 4

        # Return negative color as objective function value. The negative value is used since we want to minimize the objective function       
    return -color

def test_mkcs_G1e2c():
    ###### Trivial case with 1 edge and 2 colors,
    G1e2c = nx.Graph()
    G1e2c.add_edge(0, 1)
    num_nodes = len(G1e2c.nodes)

    color_list = ["one", "two"]

    def G1e2c_onehot():
        # ONE-HOT 
        qarg = QuantumArray(qtype = QuantumColor(color_list, one_hot_enc = True), shape = num_nodes) 

        mkcs_1e2c = QAOAProblem(create_coloring_operator(G1e2c), apply_XY_mixer, create_coloring_cl_cost_function(G1e2c))
        init_state = [random.choice(color_list) for _ in range(len(G1e2c))]

        mkcs_1e2c.set_init_function(lambda x : x.encode(init_state))
        res1e2c = mkcs_1e2c.run(qarg, depth, mes_kwargs={"backend" : qaoa_backend}, max_iter = 25)

        best_coloring, best_solution = min([(mkcs_obj(quantumcolor_array,G1e2c),quantumcolor_array) for quantumcolor_array in res1e2c.keys()], key=itemgetter(0))

            # Get final solution with optimized gamma and beta angle parameter values and print it
        best_coloring, res_str = min([(mkcs_obj(quantumcolor_array,G1e2c),quantumcolor_array) for quantumcolor_array in list(res1e2c.keys())[:5]], key=itemgetter(0))
        best_coloring, best_solution = (mkcs_obj(res_str,G1e2c),res_str)

        if res_str[0] != res_str[1]:
            return True
    
    for _ in range(5):
        if G1e2c_onehot() == True:
            break
            
    else: 
        assert False


#    onehot_simple = G1e2c_onehot()

    def G1e2c_bin():
        qarg = QuantumArray(qtype = QuantumColor(color_list, one_hot_enc = False), shape = num_nodes) 

        mkcs_1e2c = QAOAProblem(create_coloring_operator(G1e2c), RX_mixer, create_coloring_cl_cost_function(G1e2c))
        init_state = [random.choice(color_list) for _ in range(len(G1e2c))]

        mkcs_1e2c.set_init_function(lambda x : x.encode(init_state))
        res1e2c = mkcs_1e2c.run(qarg, depth, mes_kwargs={"backend" : qaoa_backend}, max_iter = 25)

        best_coloring, best_solution = min([(mkcs_obj(quantumcolor_array,G1e2c),quantumcolor_array) for quantumcolor_array in res1e2c.keys()], key=itemgetter(0))

            # Get final solution with optimized gamma and beta angle parameter values and print it
        best_coloring, res_str = min([(mkcs_obj(quantumcolor_array,G1e2c),quantumcolor_array) for quantumcolor_array in list(res1e2c.keys())[:5]], key=itemgetter(0))
        best_coloring, best_solution = (mkcs_obj(res_str,G1e2c),res_str)

        if res_str[0] != res_str[1]:
            return True
    
#    bin_simple = G1e2c_bin
    for _ in range(5):
        if G1e2c_bin() == True:
            break
    else: 
        assert False

def test_mkcs_5nodes():
    G = nx.Graph()
    G.add_edges_from([[0,1],[0,4],[1,2],[1,3],[1,4],[2,3],[3,4]])

    num_nodes = len(G.nodes)

    color_list = ["one", "two", "three", "four"]

    def G_onehot():
        qarg = QuantumArray(qtype = QuantumColor(color_list, one_hot_enc = True), shape = num_nodes) 

        mkcs_onehot = QAOAProblem(create_coloring_operator(G), apply_XY_mixer, create_coloring_cl_cost_function(G))
        init_state = [random.choice(color_list) for _ in range(len(G))]

        mkcs_onehot.set_init_function(lambda x : x.encode(init_state))

        res_onehot = mkcs_onehot.run(qarg, depth, mes_kwargs={"backend" : qaoa_backend}, max_iter = 25)

        best_coloring, best_solution = min([(mkcs_obj(quantumcolor_array,G),quantumcolor_array) for quantumcolor_array in res_onehot.keys()], key=itemgetter(0))

            # Get final solution with optimized gamma and beta angle parameter values and print it
        best_coloring_onehot, res_str_onehot = min([(mkcs_obj(quantumcolor_array,G),quantumcolor_array) for quantumcolor_array in list(res_onehot.keys())[:5]], key=itemgetter(0))
        best_coloring_onehot, best_solution_onehot = (mkcs_obj(res_str_onehot,G),res_str_onehot)
        
        if all(res_str_onehot[pair[0]] != res_str_onehot[pair[1]] for pair in list(G.edges())):
            return True


    for _ in range(5):
        if G_onehot() == True:
            break
    else: 
        assert False
#    onehot = G_onehot

    def G_bin():
        qarg = QuantumArray(qtype = QuantumColor(color_list, one_hot_enc = False), shape = num_nodes) 

        mkcs_bin = QAOAProblem(create_coloring_operator(G), RX_mixer, create_coloring_cl_cost_function(G))
        init_state = [random.choice(color_list) for _ in range(len(G))]

        mkcs_bin.set_init_function(lambda x : x.encode(init_state))

        res_bin = mkcs_bin.run(qarg, depth, mes_kwargs={"backend" : qaoa_backend}, max_iter = 25)

        best_coloring, best_solution = min([(mkcs_obj(quantumcolor_array,G),quantumcolor_array) for quantumcolor_array in res_bin.keys()], key=itemgetter(0))

        best_coloring_bin, res_str_bin = min([(mkcs_obj(quantumcolor_array,G),quantumcolor_array) for quantumcolor_array in list(res_bin.keys())[:5]], key=itemgetter(0))
        best_coloring_bin, best_solution_bin = (mkcs_obj(res_str_bin,G),res_str_bin)

        if all(res_str_bin[pair[0]] != res_str_bin[pair[1]] for pair in list(G.edges())):
            return True
    
    for _ in range(5):
        if G_bin() == True:
            break
    else: 
        assert False
#    bin = G_bin











# ONE HOT




#  BINARY




"""
\********************************************************************************
* Copyright (c) 2023 the Qrisp authors
*
* This program and the accompanying materials are made available under the
* terms of the Eclipse Public License 2.0 which is available at
* http://www.eclipse.org/legal/epl-2.0.
*
* This Source Code may also be made available under the following Secondary
* Licenses when the conditions for such availability set forth in the Eclipse
* Public License, v. 2.0 are satisfied: GNU General Public License, version 2
* with the GNU Classpath Exception which is
* available at https://www.gnu.org/software/classpath/license.html.
*
* SPDX-License-Identifier: EPL-2.0 OR GPL-2.0 WITH Classpath-exception-2.0
********************************************************************************/
"""

import numpy as np

def fourier_params_helper(array_uk_vk, fourier_depth, index_p ):
    """
    Helper function to instanciate the Optimization params via the Fourier heuristic approach, see `Zhou et al. <https://arxiv.org/pdf/1812.01041.pdf>`_  p.5

    Parameters
    ----------
    array_uk_vk : list
        initial random inits, optimization parameter
    fourier_depth : int
        depth (q) of fourier transformation
    index_p : int
        depth of (iteration of) the circuit
    """
    theta = []
    #for gamma: go through first half
    for index in range(index_p):
            # index = index*depth?
            saverlist = []
            # then go through whatever the necessary uk_s for this gamma
            for sgl_res_index in range(fourier_depth):  
                saverlist.append(array_uk_vk[sgl_res_index] * np.sin( (sgl_res_index + 1 - 1/2)*( index + 1 - 1/2)* np.pi/index_p))                     

            theta.append(sum(saverlist))

    #for beta: go through second half 
    for index2 in range(index_p):
        saverlist2 = []
        # then go through whatever the necessary vk_s for this beta 
        for sgl_res_index2 in range(fourier_depth, 2*fourier_depth):                                                              # have to subtract the depth here again
            saverlist2.append(array_uk_vk[sgl_res_index2] * np.sin( (sgl_res_index2 - fourier_depth + 1 - 1/2)*( index2 + 1 - 1/2)* np.pi/index_p))

        theta.append(sum(saverlist2))

    return theta




"""
\********************************************************************************
* Copyright (c) 2023 the Qrisp authors
*
* This program and the accompanying materials are made available under the
* terms of the Eclipse Public License 2.0 which is available at
* http://www.eclipse.org/legal/epl-2.0.
*
* This Source Code may also be made available under the following Secondary
* Licenses when the conditions for such availability set forth in the Eclipse
* Public License, v. 2.0 are satisfied: GNU General Public License, version 2
* with the GNU Classpath Exception which is
* available at https://www.gnu.org/software/classpath/license.html.
*
* SPDX-License-Identifier: EPL-2.0 OR GPL-2.0 WITH Classpath-exception-2.0
********************************************************************************/
"""

from qrisp import h, rz, rx , rzz
import itertools
import networkx as nx




def maxCliqueCostOp(G):
    
    """
    |  Based on PennyLane unconstrained mixer implementation.
    |  Initial state in :math:`(|0>+|1>)^{\otimes n}` . 
    |  For explanation see the PennyLane implementation.

    Parameters
    ----------
    G : nx.Graph
        Graph of the problem instance

    Returns
    -------
    QuantumCircuit: qrisp.QuantumCircuit
        the Operator applied to the circuit-QuantumVariable

    """
    G_compl = nx.complement(G)
    def partialcostMixer(qv, gamma):
        for pair in list(G_compl.edges()):
            #cx(qv[pair[0]], qv[pair[1]])
            rzz(3*gamma, qv[pair[0]], qv[pair[1]])
            rz(-gamma, qv[pair[0]])
            rz(-gamma, qv[pair[1]])
        rz(gamma, qv)
        #return qv

    return partialcostMixer



def maxCliqueCostfct(G):
    """

    Parameters
    ----------
    G : nx.Graph
        Graph of the problem instance

    Returns
    -------
    Costfunction : function
        the classical function for the problem instance, which takes a dictionary of measurement results as input

    """

    def aClcostFct(res_dic):
        tot_energy = 0.001
        tot_counts = 0
        for state in res_dic.keys():
            # we assume solution is right
            temp = True
            energy = 0 
            intlist = [s for s in range(len(list(state))) if list(state)[s] == "1"]
            # get all combinations of vertices in graph that are marked as |1> by the solution 
            combinations = list(itertools.combinations(intlist, 2))
            # if any combination is found in the list of G.edges(), the solution is wrong, and energy == 0
            for combination in combinations:
                if combination not in G.edges():
                    temp = False
                    break
            # else we just add the number of marked as |1> nodes
            if temp: 
                energy = -len(intlist)
            tot_energy += energy * res_dic[state]
            tot_counts += res_dic[state]

        #print(tot_energy/tot_counts)

        return tot_energy/tot_counts

    return aClcostFct



def init_state(qv):
    h(qv)
    return qv


"""
\********************************************************************************
* Copyright (c) 2023 the Qrisp authors
*
* This program and the accompanying materials are made available under the
* terms of the Eclipse Public License 2.0 which is available at
* http://www.eclipse.org/legal/epl-2.0.
*
* This Source Code may also be made available under the following Secondary
* Licenses when the conditions for such availability set forth in the Eclipse
* Public License, v. 2.0 are satisfied: GNU General Public License, version 2
* with the GNU Classpath Exception which is
* available at https://www.gnu.org/software/classpath/license.html.
*
* SPDX-License-Identifier: EPL-2.0 OR GPL-2.0 WITH Classpath-exception-2.0
********************************************************************************/
"""


from qrisp import QuantumVariable, mcx, rz, x, rx
from qrisp import control 
from _collections_abc import Iterable




def minSetCoverCostOp(sets, universe):
    """
    |  Create the cost/problem operator for this problem instance. Swapping rule is to swap a set in and out of the solution, if all elements are covered by other sets
    |  Idea - Per set: 

    * create ancillas for every element, they represent these elements
    * ancillas are set to "1" by default
    * perform multi controlled x operations on each ancilla
    * controls are given by sets with also contain the considered element
    * if all controls are "0" (see ``ctrl_state`` for ``mcx``-operation) we set this ancilla to "0"
    
    |  Then perform mcx on the qubit describing the set:
    |  If all ancillas are "1" this means that, for all elements in the list, there is already (atleast) one set in the bitstring describing the solution sets, which already contains this element. The set can then be swapped in (or out) of the solution, since all elements are covered by other sets
    |  Afterwards uncompute the ancillas

    Parameters
    ----------
    sets : list(Lists)
        The sets the universe is seperated into as by the problem definition

    universe: Tuple
        The universe for the problem instance, i.e. all possible values (all graph vertices)

    Returns
    -------
    QuantumCircuit: qrisp.QuantumCircuit
        the Operator applied to the circuit-QuantumVariable

    Examples
    --------

    Definition of the sets, given as list of lists. Full universe ``sol`` is given as a tuple
    >>> sets = [[0,7,1],[6,5],[2,3],[5,4],[8,7,0],[1]]
    >>> sol = (0,1,2,3,4,5,6,7,8)

    The relations between the sets, i.e. which vertice is in which other sets

    >>> print(get_neighbourhood_relations(sets, len_universe=len(sol)))

    Assign operators
    >>> cost_fun = minSetCoverclCostfct(sets=sets,universe = sol)
    >>> mixerOp = RZ_Mixer()
    >>> costOp = minSetCoverCostOp(sets=sets, universe=sol)
    """

    if not isinstance(sets, Iterable):
        raise Exception("Wrong structure of problem - clauses have to be iterable!")
    for clause in sets:
        if not isinstance(clause, Iterable):
            raise Exception("Wrong structure of problem - each set has to be a tuple!")
        for item in clause:
            if not isinstance(item, int):
                raise Exception("Wrong structure of problem - each literal has to an int!")

    # get neighbourhood relations from helper function
    nbh_rel = get_neighbourhood_relations(sets, len(universe))

    def theCostOpEmbedded(qv, gamma):
        # go through all sets
        for set_index in range(len(sets)):
            # get set elements and create an ancilla for every set element
            nodes = sets[set_index]
            ancillas = QuantumVariable(len(nodes))
            # go through all ancillas and, equivalently, set elements
            for ancilla_index in range(len(ancillas)):
                # if the considered element is in no other set, we cannot swap this set...
                if len(nbh_rel[nodes[ancilla_index]])<2:
                    #break ?
                    continue
                # set ancilla to one 
                x(ancillas[ancilla_index])
                # get relevant neighbourhood sets for this element
                nbh_sets_list = [item for item in nbh_rel[nodes[ancilla_index]] if item != set_index]
                # perform mcx on the ancilla, controlled by the relevant set qubits
                mcx([qv[nbh_sets_index] for nbh_sets_index in nbh_sets_list], ancillas[ancilla_index], ctrl_state= "0"*len(nbh_sets_list))
            # perform mcx on the set qubit if all elements are covered by other sets, i.e. all ancillas == "1"
            with control(ancillas):
                rx(gamma, qv[set_index])

            # uncompute ancillas
            ancillas.uncompute()
            
    return theCostOpEmbedded







def get_neighbourhood_relations(sets, len_universe):
    """
    helper function to return a dictionary describing neighbourhood relations in the sets, i.e. for each element in the universe, gives the info in which the element is contained in.


    Parameters
    ----------
    sets : list(Lists)
        The sets the universe is seperated into as by the problem definition

    len_universe: int
        The number of elements in the universe

    Returns
    -------
    neighbourhood relation dictionary :  dict
        |  keys: all universe elements (all graph vertices)
        |  values: per universe element the sets it is contained in


    """
    n_dict = {}
    for index_node in range(len_universe):
        adding_list = [index_set for index_set in range(len(sets)) if index_node in sets[index_set]]
        #if len(adding_list)>1: 
        n_dict[index_node] = adding_list
    return n_dict


def minSetCoverclCostfct(sets, universe):

    """
    create the classical cost function for the problem instance

    Parameters
    ----------
    sets : list(Lists)
        The sets the universe is seperated into as by the problem definition
        
    universe: Tuple
        The universe for the problem instance, i.e. all possible values (all graph vertices)

    Returns
    -------
    Costfunction : function
        the classical function for the problem instance, which takes a dictionary of measurement results as input

    """
    def setupaClCostfct(res_dic):
        energy = 0
        total_counts = 0
        for state in list(res_dic.keys()):
            obj = 0.01
            intlist = [s for s in range(len(list(state))) if list(state)[s] == "1"]
            sol_sets = [sets[index] for index in intlist]
            res = ()
            # get all elements that are contained within the solution sets
            for seto in sol_sets:
                res = tuple(set(res + tuple(seto))) 
            # this is problematic... wrong solutions are being disregarded...
            if res  != universe:
                continue
            obj += len(intlist)
            energy += obj * res_dic[state]
            total_counts += res_dic[state]
        #print(energy/total_counts)

        return energy/total_counts
    
    return setupaClCostfct



def init_state(qv):
    x(qv)
    return qv






"""
\********************************************************************************
* Copyright (c) 2023 the Qrisp authors
*
* This program and the accompanying materials are made available under the
* terms of the Eclipse Public License 2.0 which is available at
* http://www.eclipse.org/legal/epl-2.0.
*
* This Source Code may also be made available under the following Secondary
* Licenses when the conditions for such availability set forth in the Eclipse
* Public License, v. 2.0 are satisfied: GNU General Public License, version 2
* with the GNU Classpath Exception which is
* available at https://www.gnu.org/software/classpath/license.html.
*
* SPDX-License-Identifier: EPL-2.0 OR GPL-2.0 WITH Classpath-exception-2.0
********************************************************************************/
"""

from qrisp.qaoa import QAOAProblem
from qrisp.qaoa.problems.maxSetPackInfrastr import maxSetPackclCostfct,maxSetPackCostOp,get_neighbourhood_relations, init_state
from qrisp.qaoa.mixers import RZ_mixer
from qrisp import QuantumVariable
from qrisp.qaoa import approximation_ratio
import matplotlib.pyplot as plt
import itertools


def QAOAmaxSetPacking():
    # start with a graph G - from np adjacency matrix
    """
    This is an example of the maxSetPacking QAOA solution
    the problem is structured as follows:
    > Given a universe [n] and m subsets S (S_j in S, S_j is subset of [n]), find the maximum cardinality subcollection S' (S' is subset of S) of pairwise dijoint sets.

    We will not stick to mathematical assignment of variable names. 

    """

    # sets are given as list of tuples
    examplesets = [(0,1,7),(5,6),(2,3),(4,5),(7,8)]
    # full universe is given as a tuple
    sol = (0,1,2,3,4,5,6,7,8)

    # the realtions between the sets, i.e. with vertice is in which other sets
    #print(get_neighbourhood_relations(examplesets, len_universe=len(sol)))

    # assign the operators
    cost_fun = maxSetPackclCostfct(sets=examplesets,universe=sol)
    mixerOp = RZ_mixer
    costOp = maxSetPackCostOp(sets=examplesets, universe=sol)

    #initialize the qarg
    qarg = QuantumVariable(len(examplesets))

    # run the qaoa
    QAOAinstanceSecond = QAOAProblem(cost_operator=costOp ,mixer= mixerOp, cl_cost_function=cost_fun)
    QAOAinstanceSecond.set_init_function(init_function=init_state)
    InitTest = QAOAinstanceSecond.run(qarg=qarg, depth=5)


    # assign the cost_func for later usage
    def testCostFun(state, universe):
        list_universe = [True]*len(universe)
        temp = True
        obj = 0
        intlist = [s for s in range(len(list(state))) if list(state)[s] == "1"]
        sol_sets = [examplesets[index] for index in intlist]
        for seto in sol_sets:
            for val in seto:
                if list_universe[val]:
                    list_universe[val] = False
                else: 
                    temp = False 
                    break
        if temp:
            obj -= len(intlist)
        return obj




    # test if cost_func functions for all items . 
    for state in InitTest.keys():
        list_universe = [True]*len(sol)
        temp = True
        if testCostFun(state=state, universe = sol) < 0:
            intlist = [s for s in range(len(list(state))) if list(state)[s] == "1"]
            sol_sets = [examplesets[index] for index in intlist]
            for seto in sol_sets:
                for val in seto:
                    if list_universe[val]:
                        list_universe[val] = False
                    else: 
                        temp = False 
                        break
        assert temp

    

    #print the ideal solutions
    #print("5 most likely Solutions") 
    maxfive = sorted(InitTest, key=InitTest.get, reverse=True)[:5]
    """ for name in InitTest.keys():  # for name, age in dictionary.iteritems():  (for Python 2.x)
        if name in maxfive:
            print(name)
            print(testCostFun(name, universe=sol))  
    """
    #find ideal solution by brute force    
    temp_binStrings = list(itertools.product([1,0], repeat=len(examplesets)))
    binStrings = ["".join(map(str, item))  for item in temp_binStrings]
    
    mini = 0
    for index in range(len(binStrings)):
        val = testCostFun(binStrings[index], universe=sol)
        if val < mini:
            mini = val
            min_index = index

    optimal_sol = binStrings[min_index]
    #print("optimal solution")
    #print(optimal_sol)


    # write approximation ratio test
    metric_dict ={}
    metric_dict["counts"] = InitTest
    metric_dict["cost_func"] = maxSetPackclCostfct(sets=examplesets,universe=sol)
    metric_dict["optimal_sol"] = optimal_sol
    
    #print(approximation_ratio(metric_dict))
    """ if not approximation_ratio(metric_dict)>= 0.3:
        failure_list.append(approximation_ratio(metric_dict))
        counter +=1 """
    #assert approximation_ratio(metric_dict)>= 0.3



    import numpy as np

    randInt = np.random.randint(4,9)
    randInt2 = np.random.randint(0,1)
    secondSol = tuple(range(randInt))

    if randInt2 == 0:
        secondSets = [secondSol,secondSol[0:randInt-3],secondSol[randInt-3:randInt-1],secondSol[randInt-1:]] 
    else:
        #print(list(range(randInt)))
        #examplesets = [tuple(index) for index in list(range(randInt))]
        secondSets = [secondSol[0:randInt-3],secondSol[randInt-3:randInt-1],secondSol[randInt-1:]] 



    # assign the operators
    cost_fun = maxSetPackclCostfct(sets=secondSets,universe=secondSol)
    mixerOp = RZ_mixer
    costOp = maxSetPackCostOp(sets=secondSets, universe=secondSol)

    #initialize the qarg
    qarg2 = QuantumVariable(len(secondSets))

    # run the qaoa
    QAOAinstanceSecond = QAOAProblem(cost_operator=costOp ,mixer= mixerOp, cl_cost_function=cost_fun)
    QAOAinstanceSecond.set_init_function(init_function=init_state)
    InitTest2 = QAOAinstanceSecond.run(qarg=qarg2, depth=5)

    maxOne =  sorted(InitTest2, key=InitTest2.get, reverse=True)[:1]
    if randInt2 == 0:
        
        #print(maxOne)
        assert maxOne == ["0111"]

    else:
        #print(maxOne)
        assert maxOne == ["1"* 3 ]
    #print(counter)
    #print(failure_list)


"""
\********************************************************************************
* Copyright (c) 2023 the Qrisp authors
*
* This program and the accompanying materials are made available under the
* terms of the Eclipse Public License 2.0 which is available at
* http://www.eclipse.org/legal/epl-2.0.
*
* This Source Code may also be made available under the following Secondary
* Licenses when the conditions for such availability set forth in the Eclipse
* Public License, v. 2.0 are satisfied: GNU General Public License, version 2
* with the GNU Classpath Exception which is
* available at https://www.gnu.org/software/classpath/license.html.
*
* SPDX-License-Identifier: EPL-2.0 OR GPL-2.0 WITH Classpath-exception-2.0
********************************************************************************/
"""



from qrisp.qaoa.parameters.fourier_params import fourier_params_helper
def fourier_optimization_wrapper(array_uk_vk, qc, symbols, qarg_dupl , cl_cost_function, fourier_depth, index_p ,mes_kwargs):
        """
        Wrapper function for the optimization method used in QAOA, based on the Fourier heuristic.

        This function calculates the value of the classical cost function after post-processing if a post-processing function is set, otherwise it calculates the value of the classical cost function.

        Parameters
        ----------
        array_uk_vk : list
            The parameters to be optimized
        qc : QuantumCircuit
            The compiled quantum circuit.
        symbols : list
            The list of symbols used in the quantum circuit.
        qarg_dupl : QuantumVariable
            The duplicated quantum variable to which the quantum circuit is applied.
        mes_kwargs : dict
            The keyword arguments for the measurement function.

        Returns
        -------
        float
            The value of the classical cost function.
        """

        theta = fourier_params_helper(array_uk_vk, fourier_depth ,index_p)

        subs_dic = {symbols[i] : theta[i] for i in range(len(symbols))}
        res_dic = qarg_dupl.get_measurement(subs_dic = subs_dic, precompiled_qc = qc, **mes_kwargs)
        cl_cost = cl_cost_function(res_dic)
        

        return cl_cost


"""
\********************************************************************************
* Copyright (c) 2023 the Qrisp authors
*
* This program and the accompanying materials are made available under the
* terms of the Eclipse Public License 2.0 which is available at
* http://www.eclipse.org/legal/epl-2.0.
*
* This Source Code may also be made available under the following Secondary
* Licenses when the conditions for such availability set forth in the Eclipse
* Public License, v. 2.0 are satisfied: GNU General Public License, version 2
* with the GNU Classpath Exception which is
* available at https://www.gnu.org/software/classpath/license.html.
*
* SPDX-License-Identifier: EPL-2.0 OR GPL-2.0 WITH Classpath-exception-2.0
********************************************************************************/
"""

import numpy as np
from sympy import continued_fraction_convergents, continued_fraction_iterator, Rational

from qrisp.interface import VirtualQiskitBackend
from qrisp.arithmetic.modular_arithmetic import find_optimal_m, modinv
from qrisp import QuantumModulus, QuantumFloat, h, control, QFT

depths = []
cnot_count = []
qubits = []

def find_optimal_a(N):
    n = int(np.ceil(np.log2(N)))
    proposals = []
    
    # Search through the first O(1) possibilities to find a good a
    for a in range(2, min(100, N-1)):
        # We only append non-trivial proposals
        if np.gcd(a, N) == 1:
            proposals.append(a)
    
    cost_dic = {}
    for a in proposals:
        m_values = []
        for k in range(2*n+1):
            inpl_multiplier = (a**(2**k))%N
            
            if inpl_multiplier == 1:
                continue
            
            # find_optimal_m is a function that determines the lowest possible
            # Montgomery shift for a given number. The higher the montgomery shift,
            # the more qubits and the more effort is needed.
            m_values.append(find_optimal_m(inpl_multiplier, N))
            m_values.append(find_optimal_m(modinv((-inpl_multiplier)%N, N), N))
        
        cost_dic[a] = sum(m_values) + max(m_values)*1E-5
        
    proposals.sort(key = lambda a : cost_dic[a])
    
    
    optimal_a = proposals[0]
    
    m_values = []
    
    for k in range(2*n+1):
        inpl_multiplier = ((optimal_a)**(2**k))%N
        
        if inpl_multiplier == 1:
            continue
        
        m_values.append(find_optimal_m(inpl_multiplier, N))
    
    return proposals

def find_order(a, N, inpl_adder = None, mes_kwargs = {}):
    qg = QuantumModulus(N, inpl_adder)
    qg[:] = 1
    qpe_res = QuantumFloat(2*qg.size + 1, exponent = -(2*qg.size + 1))
    h(qpe_res)
    for i in range(len(qpe_res)):
        with control(qpe_res[i]):
            qg *= a
            a = (a*a)%N
    QFT(qpe_res, inv = True, inpl_adder = inpl_adder)
    
    mes_res = qpe_res.get_measurement(**mes_kwargs)
    
    return extract_order(mes_res, a, N)


def extract_order(mes_res, a, N):
    
    collected_r_values = []
    
    approximations = list(mes_res.keys())
    
    try:
        approximations.remove(0)
    except ValueError:
        pass
    
    while True:
        
        r_values = get_r_values(approximations.pop(0))
        
        for r in r_values:  
            if (a**r)%N == 1:
                return r
        
        collected_r_values.append(r_values)
        from itertools import product
        
        for comb in product(*collected_r_values):
            r = np.lcm.reduce(comb)
            if (a**r)%N == 1:
                return r
    
def get_r_values(approx):
    rationals = continued_fraction_convergents(continued_fraction_iterator(Rational(approx)))
    return [rat.q for rat in rationals if 1 < rat.q]
    

def shors_alg(N, inpl_adder = None, mes_kwargs = {}):
    """
    Performs `Shor's factorization algorithm <https://arxiv.org/abs/quant-ph/9508027>`_ on a given integer N.
    The adder used for factorization can be customized. To learn more about this feature, please read :ref:`QuantumModulus`

    Parameters
    ----------
    N : integer
        The integer to be factored.
    inpl_adder : callable, optional
        A function that performs in-place addition. The default is None.
    mes_kwargs : dict, optional
        A dictionary of keyword arguments for :meth:`get_measurement <qrisp.QuantumVariable.get_measurement>`. This especially allows you to specify an execution backend. The default is {}.

    Returns
    -------
    res : integer
        A factor of N.

    Examples
    --------
    
    We factor 65:
        
    >>> from qrisp.shor import shors_alg
    >>> shors_alg(65)
    5

    """
    if not N%2:
        return 2
    
    a_proposals = find_optimal_a(N)
    
    for a in a_proposals:
        
        K = np.gcd(a, N)
        
        if K != 1:
            res = K
            break
        
        r = find_order(a, N, inpl_adder, mes_kwargs)
        
        if r%2:
            continue
        
        g = int(np.gcd(a**(r//2)+1, N))
        
        if g not in[N, 1]:
            res = g
            break
    return res


"""
\********************************************************************************
* Copyright (c) 2023 the Qrisp authors
*
* This program and the accompanying materials are made available under the
* terms of the Eclipse Public License 2.0 which is available at
* http://www.eclipse.org/legal/epl-2.0.
*
* This Source Code may also be made available under the following Secondary
* Licenses when the conditions for such availability set forth in the Eclipse
* Public License, v. 2.0 are satisfied: GNU General Public License, version 2
* with the GNU Classpath Exception which is
* available at https://www.gnu.org/software/classpath/license.html.
*
* SPDX-License-Identifier: EPL-2.0 OR GPL-2.0 WITH Classpath-exception-2.0
********************************************************************************/
"""


from qrisp.qaoa import QAOAProblem
from qrisp.qaoa.problems.maxSatInfrastr import maxSatclCostfct, maxSatCostOp, init_state, clausesdecoder
from qrisp.qaoa.mixers import RX_mixer
from qrisp import QuantumVariable
import networkx as nx
import matplotlib.pyplot as plt


def test_QAOAmaxSat():

    #draw graph
    #nx.draw(giraf,with_labels = True)
    #plt.show() 

    problem = [6, [[1,2,-3],[1,4,-6], [4,5,6],[1,3,-4],[2,4,5],[1,3,5],[-2,-3,6]]]
    clauses11 = problem[1]

    #Clauses are decoded, s.t. the Cost-Optimizer can read them
    #numVars is the amount of considered variables, i.e. highest number (= Number of Qubits in Circuit aswell)
    decodedClauses = clausesdecoder( problem)
    #print(decodedClauses)

    qarg = QuantumVariable(6)

    #CostOperator-Generator has to be called with the clauses
    #CostFct-Generator has to be called with decodedClauses
    QAOAinstance = QAOAProblem(cost_operator=maxSatCostOp(problem), mixer=RX_mixer, cl_cost_function=maxSatclCostfct(problem))
    QAOAinstance.set_init_function(init_function=init_state)
    theNiceQAOA = QAOAinstance.run(qarg=qarg, depth=5)


    import itertools
    def testCostFun(state):
        obj = 0
        #literally just do a minus 1 op if state is equiv to a given condition
        for index in range(len(decodedClauses)):
            if state in decodedClauses[index]:
                obj -= 1 

        return obj

    
    #print 5 best sols 
    # assert that sol is in decoded clauses 
    maxfive = sorted(theNiceQAOA, key=theNiceQAOA.get, reverse=True)[:5]
    # assert that sol is in decoded clauses 
    for name in theNiceQAOA.keys():  # for name, age in dictionary.iteritems():  (for Python 2.x)
        if name in maxfive:
            print(name)
            print(testCostFun(name))
        if testCostFun(name) < 0:
            temp = False
            for index in range(len(decodedClauses)):
                if name in decodedClauses[index]:
                    temp  = True
            assert temp




"""
\********************************************************************************
* Copyright (c) 2023 the Qrisp authors
*
* This program and the accompanying materials are made available under the
* terms of the Eclipse Public License 2.0 which is available at
* http://www.eclipse.org/legal/epl-2.0.
*
* This Source Code may also be made available under the following Secondary
* Licenses when the conditions for such availability set forth in the Eclipse
* Public License, v. 2.0 are satisfied: GNU General Public License, version 2
* with the GNU Classpath Exception which is
* available at https://www.gnu.org/software/classpath/license.html.
*
* SPDX-License-Identifier: EPL-2.0 OR GPL-2.0 WITH Classpath-exception-2.0
********************************************************************************/
"""

from qrisp.qaoa import *
import networkx as nx
from operator import itemgetter

def test_QUBO():
    from qrisp.default_backend import def_backend

    Q1 = np.array(
        [
            [1,0,0,0],
            [0,0,-0.5,0.5],
            [0,-0.5,0,1],
            [0,0.5,1,-2],
        ]
    )
    qarg = QuantumArray(qtype = QuantumVariable(1), shape = len(Q1))
    QUBO_instance = QUBO_problem(Q1)
    res = QUBO_instance.run(qarg, depth=1, mes_kwargs={"backend" : def_backend}, max_iter = 50)
    best_cost1, res_str = min([(QUBO_obj(bitstring, Q1), bitstring) for bitstring in list(res.keys())], key=itemgetter(0))

    assert best_cost1 == -2.0

    Q2 = np.array(
        [
            [1,-3,-3,-3],
            [-3,1,0,0],
            [-3,0,1,-3],
            [-3,0,-3,1],
        ]
    )
    
    qarg = QuantumArray(qtype = QuantumVariable(1), shape = len(Q2))
    QUBO_instance = QUBO_problem(Q2)
    res = QUBO_instance.run(qarg, depth=1, mes_kwargs={"backend" : def_backend}, max_iter = 50)
    best_cost2, res_str = max([(QUBO_obj(bitstring, Q2), bitstring) for bitstring in list(res.keys())], key=itemgetter(0))

    assert best_cost2 == 2

    Q3 = np.array(
        [
            [1,-6,-6,-6],
            [0,1,0,0],
            [0,0,1,-6],
            [0,0,0,1],
        ]
    )

    qarg = QuantumArray(qtype = QuantumVariable(1), shape = len(Q3))
    QUBO_instance = QUBO_problem(Q3)
    res = QUBO_instance.run(qarg, depth=1, mes_kwargs={"backend" : def_backend}, max_iter = 50)
    best_cost3, res_str = max([(QUBO_obj(bitstring, Q2), bitstring) for bitstring in list(res.keys())], key=itemgetter(0))

    assert best_cost3 == best_cost2
    



from qulacs import QuantumState, QuantumCircuit
import qulacs.gate as gt


def qulacs_converter(qc):
    from qrisp.simulator import extract_measurements

    temp_qc, measurements = extract_measurements(qc.copy())

    transpiled_qc = temp_qc.transpile(basis_gates = ["cx", "x", "y", "z", "h", "rz", "ry", "rx", "swap" ," id" , "mcx", "S", "Sdg", "t"])
    
    qulacs_qc = QuantumCircuit(len(qc.qubits))
    
    for instr in transpiled_qc.data:
        if instr.op.name == "x":
            qulacs_qc.add_gate(gt.X(transpiled_qc.qubits.index(instr.qubits[0])))
            
        elif instr.op.name == "y":
            qulacs_qc.add_gate(gt.Y(transpiled_qc.qubits.index(instr.qubits[0])))

        elif instr.op.name == "z":
            qulacs_qc.add_gate(gt.Z(transpiled_qc.qubits.index(instr.qubits[0])))

        elif instr.op.name == "h":
            qulacs_qc.add_gate(gt.H(transpiled_qc.qubits.index(instr.qubits[0])))
        
        elif instr.op.name == "cx":
            qulacs_qc.add_gate(gt.CNOT(transpiled_qc.qubits.index(instr.qubits[0]), transpiled_qc.qubits.index(instr.qubits[1])))
    
        elif instr.op.name == "rz":
            qulacs_qc.add_gate(gt.RZ(transpiled_qc.qubits.index(instr.qubits[0]), instr.op.params[0]))

        elif instr.op.name == "ry":
            qulacs_qc.add_gate(gt.RY(transpiled_qc.qubits.index(instr.qubits[0]), instr.op.params[0]))

        elif instr.op.name == "rx":
            qulacs_qc.add_gate(gt.RX(transpiled_qc.qubits.index(instr.qubits[0]), instr.op.params[0]))

        elif instr.op.name == "swap":
            qulacs_qc.add_gate(gt.SWAP(transpiled_qc.qubits.index(instr.qubits[0])))

        elif instr.op.name == "id":
            qulacs_qc.add_gate(gt.Identity(transpiled_qc.qubits.index(instr.qubits[0])))

        #elif instr.op.name == "mcx":
            #qulacs_qc.add_gate(gt.TOFFOLI(transpiled_qc.qubits.index(instr.qubits[0]), transpiled_qc.qubits.index(instr.qubits[1]))) 
        
        elif instr.op.name == "S":
            qulacs_qc.add_gate(gt.S(transpiled_qc.qubits.index(instr.qubits[0])))

        elif instr.op.name == "Sdg":
            qulacs_qc.add_gate(gt.Sdag(transpiled_qc.qubits.index(instr.qubits[0])))

        elif instr.op.name == "t":
            qulacs_qc.add_gate(gt.T(transpiled_qc.qubits.index(instr.qubits[0])))
            
        else:
            raise Exception(f"Don't know how to convert gate {instr.op.name}")
        
    return qulacs_qc


"""
\********************************************************************************
* Copyright (c) 2023 the Qrisp authors
*
* This program and the accompanying materials are made available under the
* terms of the Eclipse Public License 2.0 which is available at
* http://www.eclipse.org/legal/epl-2.0.
*
* This Source Code may also be made available under the following Secondary
* Licenses when the conditions for such availability set forth in the Eclipse
* Public License, v. 2.0 are satisfied: GNU General Public License, version 2
* with the GNU Classpath Exception which is
* available at https://www.gnu.org/software/classpath/license.html.
*
* SPDX-License-Identifier: EPL-2.0 OR GPL-2.0 WITH Classpath-exception-2.0
********************************************************************************/
"""

import numpy as np

from qrisp import QuantumCircuit, Operation, QuantumVariable
from qrisp.mcx_algs.circuit_library import jones_toffoli_qc, amy_toffoli_qc, ctrl_state_wrap

# Implementation of the T-Depth 1 Toffoli found in https://arxiv.org/pdf/1212.5069.pdf
class JonesToffoli(Operation):
    
    def __init__(self, ctrl_state = "11"):
        
        
        definition = QuantumCircuit(5)
        
        if ctrl_state[0] == "0":
            definition.x(definition.qubits[0])
        if ctrl_state[1] == "0":
            definition.x(definition.qubits[1])
        
        definition.append(amy_toffoli_qc.to_gate("jones_toffoli"), definition.qubits[:-1])
        
        if ctrl_state[0] == "0":
            definition.x(definition.qubits[0])
        if ctrl_state[1] == "0":
            definition.x(definition.qubits[1])
        
        name = "uncompiled_jones_toffoli"
        
        Operation.__init__(self, name = name, num_qubits = 5, definition = definition)
        
        self.permeability = {0: True, 1: True, 2: False, 3 : False, 4 : False}
        self.is_qfree = True
        self.ctrl_state = ctrl_state
    
    def recompile(self):
        
        res = ctrl_state_wrap(jones_toffoli_qc, self.ctrl_state).to_op("compiled_jones_toffoli")
        res.permeability = {0: True, 1: True, 2: False, 3 : False, 4 : False}
        res.is_qfree = True
            
        return res
    
    def inverse(self):
        return JonesToffoli(self.ctrl_state)
    
def jones_toffoli(ctrl, target, ctrl_state = "11"):
    if isinstance(target, (list, QuantumVariable)):
        target = target[0]
    
    jones_ancilla = QuantumVariable(2, qs = ctrl[0].qs(), name = "jones_ancilla_" + str(int(QuantumVariable.creation_counter[0])))
    
    target.qs().append(JonesToffoli(ctrl_state), list(ctrl) + [target] + list(jones_ancilla))
    jones_ancilla.delete()


"""
\********************************************************************************
* Copyright (c) 2023 the Qrisp authors
*
* This program and the accompanying materials are made available under the
* terms of the Eclipse Public License 2.0 which is available at
* http://www.eclipse.org/legal/epl-2.0.
*
* This Source Code may also be made available under the following Secondary
* Licenses when the conditions for such availability set forth in the Eclipse
* Public License, v. 2.0 are satisfied: GNU General Public License, version 2
* with the GNU Classpath Exception which is
* available at https://www.gnu.org/software/classpath/license.html.
*
* SPDX-License-Identifier: EPL-2.0 OR GPL-2.0 WITH Classpath-exception-2.0
********************************************************************************/
"""

import numpy as np

from qrisp.circuit import QuantumCircuit
from qrisp.misc.GMS_tools import GXX_wrapper


# Function to synthesize a multi controlled CX gate from GMS gates
def gms_multi_cx(n):
    qc = QuantumCircuit(n + 1)

    if n == 2:
        for i in range(3):
            qc.ry(np.pi / 2, i)

        qc.rz(np.pi / 4, 2)

        qc.append(GXX_wrapper(3, 3 * [3 * [np.pi / 2]]), qc.qubits)

        for i in range(3):
            qc.rx(-np.pi / 2, i)

        qc.rz(-np.pi / 2, 2)

        for i in range(3):
            qc.rx(-np.pi / 4, i)

        qc.append(GXX_wrapper(3, 3 * [3 * [np.pi / 4]]), qc.qubits)

        qc.rz(np.pi / 2, 2)

        qc.append(GXX_wrapper(3, 3 * [3 * [np.pi / 2]]), qc.qubits)

        for i in range(3):
            qc.rx(np.pi / 2, i)
            qc.ry(-np.pi / 2, i)

        result = qc.to_gate()

        result.name = "GMS toffoli"

        return result
    else:
        raise Exception(str(n) + "-controlled x gate not implemented for gms method")



from qrisp import rz, rzz,rx, mcx, x, QuantumVariable, control
import numpy as np
import copy
from qrisp.qiro.qiroproblems.qiro_utils import * 
from qrisp.qaoa.problems.maxSetPackInfrastr import get_neighbourhood_relations
import networkx as nx


def trafo_maxPackToMIS(problem):
    """
    Function to transform a Maximum Set Packing problem into a Maximum independent set problem.

    Parameters
    ----------
    problem : List
        The problem definition in analogy to the QAOA maxSetPacking problem

    Returns
    -------
    G : nx.Graph
        The corresponding graph to be solved by an MIS implementation.


    """
    # the MIS solution to G is equivalent to the solution of the maxSetPackingProblem
    G = nx.Graph()
    G.add_nodes_from(list(range(len(problem[1]))))
    n_dict = get_neighbourhood_relations(problem)
    
    for val in n_dict.values():
        
        if len(val)<2:
            continue
        for index1 in range(len(val)-1):
            for index2 in range(index1+1, len(val)):
                G.add_edge(val[index1],val[index2])

    return G 










"""
does all of the set popping actually work? Since the amount of sets is equivalent to len of qarg?
--> setting to empty also doesnt work, since that implies that that they are always in

    # what to do about sols and exclusions in cost_op

where does the neighbourhood_dict fit it?

need to write cost function?


"""

""" def create_maxSetPack_replacement_routine( res, old_n_dict, problem, solutions, exclusions):

    sets ,universe = copy.deepcopy(problem[0]), problem[1]

    n_dict = copy.deepcopy(old_n_dict)

    # FOR SINGLE QUBIT CORRELATIONS
    orig_nodes = list(universe)
    combinations = []

    # add check for solution nodes      
    for index1 in range(len(universe)):
        for index2 in range(index1,len(universe)):
            combinations.append((index1,index2))

    # you are HERE!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
    max = 0
    max_item = []
    #get the max_edge and eval the sum and sign


    max_item, sign = find_max(orig_nodes, combinations, res, solutions)
    
    # we just directly remove vertices from the graph 
    if isinstance(max_item, int):
        if sign > 0:
            the_set = sets[max_item]
            for node in the_set:
                rels = n_dict[node]
                for rel in rels:
                    sets[rels] = [] 
            # add set to sol
            # go through n_dict and remove all other sets that contain elements in this set
            # what does remove mean in this context
            solutions.append(max_item)
            exclusions.append(max_item)

        elif sign < 0:
            # remove set from sets?
            sets[rels] = [] 
            exclusions.append(max_item)

    
    else:
        if sign > 0:
            
            
            # both are in: all elements can be deleted for other sets 
            # both are out: ??
            # --> add whatever is in second set to first set, and then dont consider the second set anymore?
            # keep the correlation somehow?? --> this is now in the cost_op
            # or work on the neighbourhood_dict?
            sets[0] += sets[1]
            sets[1] = []
            solutions.append(max_item)

            pass

        elif sign < 0:
            # remove all sets that contain elements of the schnittmenge, because one of the two is going to be in there
            # how to fix the negative correlation??
            intersect = list(sets[max_item[0]] & sets[max_item[1]])
            for item in intersect:
                rels = n_dict[item]
                for rel in rels: 
                    if not rel in max_item:
                        sets[rels] = [] 
            exclusions.append(rel)
            


    return sets, solutions, sign, exclusions

# what about other operatores??



def maxSetPackCost_reduced(problem, solutions = []):
    
    |  Create the cost/problem operator for this problem instance. The swapping rule is to swap a set in and out of the solution, if it is not intersecting with any other set.
    |  Idea - Per set: 

    * Create ancillas for every element, they represent these elements
    * Perform multi controlled x operations on each ancilla
    * Controls are given by sets with also contain the considered element
    * If all controls are "0" (see ``ctrl_state`` for ``mcx``-operation) we set this ancilla to "1"

    |  Then perform mcx on the qubit describing the set as follows:
    |  If all ancillas are "1" this means the qubits describing the sets contain no intersections with the considered set. We can then swap the set in (or out).
    |  Afterwards uncompute the ancillas.

    Parameters
    ----------
    sets : list(Lists)
        The sets the universe is seperated into as by the problem definition

    universe: Tuple
        The universe for the problem instance, i.e. all possible values (all graph vertices)

    Returns
    -------
    QuantumCircuit: qrisp.QuantumCircuit
        the Operator applied to the circuit-QuantumVariable



    universe, sets = list(range(problem[0])), problem[1]

    # get neigbhourhood relations from helper function
    nbh_rel = get_neighbourhood_relations(problem)

    # what to do about sols and exclusions
    def theCostOpEmbedded(qv, gamma):
        #check all sets
        for set_index in range(len(sets)):
            # get set elements and create an ancilla for every set element
            nodes = sets[set_index]
            ancillas = QuantumVariable(len(nodes))
            # go through all ancillas and, equivalently set elements
            for ancilla_index in range(len(ancillas)):
                # if the element is only in one set, we can set this ancilla to 1
                if len(nbh_rel[nodes[ancilla_index]])<2:
                    x(ancillas[ancilla_index])
                    continue
                # else save the sets with also contain the considered element
                nbh_sets_list = [ item for item in nbh_rel[nodes[ancilla_index]] if item != set_index]
                # perform mcx on ancilla, control given by the relevant set
                mcx([qv[nbh_sets_index] for nbh_sets_index in nbh_sets_list], ancillas[ancilla_index], ctrl_state= "0" * len(nbh_sets_list))
            # perform mcrx gate on the qubit describing the considered set
            with control(ancillas):
                rx(gamma, qv[set_index])  
            
            for item in solutions:
                if not isinstance(item, int):
                    cx(item[0],item[1])
            #mcrx_gate = RXGate(gamma).control(len(ancillas))
            #qv.qs.append(  mcrx_gate, [*ancillas, qv[set_index]])

            ancillas.uncompute()

    return theCostOpEmbedded

 """


"""
\********************************************************************************
* Copyright (c) 2023 the Qrisp authors
*
* This program and the accompanying materials are made available under the
* terms of the Eclipse Public License 2.0 which is available at
* http://www.eclipse.org/legal/epl-2.0.
*
* This Source Code may also be made available under the following Secondary
* Licenses when the conditions for such availability set forth in the Eclipse
* Public License, v. 2.0 are satisfied: GNU General Public License, version 2
* with the GNU Classpath Exception which is
* available at https://www.gnu.org/software/classpath/license.html.
*
* SPDX-License-Identifier: EPL-2.0 OR GPL-2.0 WITH Classpath-exception-2.0
********************************************************************************/
"""


from qrisp.qaoa import QAOAProblem
from qrisp.qaoa.problems.maxCliqueInfrastr import maxCliqueCostfct,maxCliqueCostOp,init_state
from qrisp.qaoa.mixers import RX_mixer
from qrisp import QuantumVariable
import networkx as nx
import matplotlib.pyplot as plt
import  numpy as np


def test_QAOAmaxClique():

    giraf = nx.erdos_renyi_graph(9,0.7)
    #draw graph
    #nx.draw(giraf,with_labels = True)
    #plt.show() 


    qarg = QuantumVariable(giraf.number_of_nodes())



    #Instanciate QAOA
    #mixer gets Graph as argument
    #cost function gets graph as argument 
    QAOAinstance = QAOAProblem(cost_operator= maxCliqueCostOp(giraf), mixer= RX_mixer, cl_cost_function=maxCliqueCostfct(giraf)) 
    QAOAinstance.set_init_function(init_function=init_state)
    theNiceQAOA = QAOAinstance.run(qarg=qarg, depth= 5)

    import itertools
    def aClcostFct(state, G):
        # we assume solution is right
        temp = True
        energy = 0 
        intlist = [s for s in range(len(list(state))) if list(state)[s] == "1"]
        # get all combinations of vertices in graph that are marked as |1> by the solution 
        combinations = list(itertools.combinations(intlist, 2))
        # if any combination is not found in the list of G.edges(), the solution is wrong, and energy == 0
        for combination in combinations:
            if combination not in G.edges():
                temp = False
        # else we just add the number of marked as |1> nodes
        if temp: 
            energy = -len(intlist)
        return(energy)

    
    # find the nx solutions
    the_it = list(nx.find_cliques(giraf))
    the_itt = [set(it) for it in the_it]
    print(the_itt)
    
    #print the ideal solutions
    #print("5 most likely Solutions") 
    maxfive = sorted(theNiceQAOA, key=theNiceQAOA.get, reverse=True)[:5]
    print(maxfive)
    for name in theNiceQAOA.keys():  # for name, age in dictionary.iteritems():  (for Python 2.x)
        if name in maxfive:
            print(name)
            print(aClcostFct(name, giraf))
        if aClcostFct(name, giraf) < 0:
            intlist = [s for s in range(len(list(name))) if list(name)[s] == "1"]
            intlist2 = set(intlist)
            set_g = False
            for seto in the_itt:
                if intlist2.issubset(seto):
                    set_g = True
                    break
            assert set_g

                #assert set(intlist) in the_itt
            #assert aClcostFct(name, giraf) <= -1

    # test for full solution...
    truth_list = []
    for index in range(10):

        integ = np.random.randint(1,7)
        #create random graph with either all nodes connected, or no nodes connected
        integ2 = np.random.randint(0,1)
        
        secondGraph = nx.erdos_renyi_graph(integ,integ2)
        secondQarg = QuantumVariable(secondGraph.number_of_nodes())
        secondInstance = QAOAProblem(cost_operator= maxCliqueCostOp(secondGraph), mixer= RX_mixer, cl_cost_function=maxCliqueCostfct(secondGraph)) 
        secondInstance.set_init_function(init_function=init_state)
        theNiceQAOA2 = secondInstance.run(qarg=secondQarg, depth= 3, max_iter = 50, mes_kwargs = {"shots" : 100000})
        maxOne = sorted(theNiceQAOA2, key=theNiceQAOA2.get, reverse=True)[:1]
        #if no nodes connected, maxClique sol should just be the first node, i.e. "1000" (more or less zeros)
        if integ2 == 0:
            string = "1"
            string += str(integ2) * (secondGraph.number_of_nodes()-1)
            testStr = [string]
        #if all nodes connected, maxClique sol should be all nodes, i.e. "1111" (more or less ones)
        if integ2 == 1:
            testStr = [str(integ2) * secondGraph.number_of_nodes()]

        
        if  testStr == maxOne:
            truth_list.append(1)
            
    assert sum(truth_list)/10 > 0.5
